from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy
import inspect
from pandas.core import algorithms
import pandas as pd


# Custom SeriesGroupBy class ==================================================

class GroupByAgg(SeriesGroupBy):
    def __init__(self, *args, orig_grouper, orig_obj, should_cast, **kwargs):
        self._orig_grouper = orig_grouper
        self._orig_obj = orig_obj
        self._should_cast = should_cast
        super().__init__(*args, **kwargs)
    
    def _broadcast_agg_result(self):
        return broadcast_agg_result(
            self._orig_grouper, 
            self.obj,
            self._orig_obj,
            cast = self._should_cast
        )
    
    @classmethod
    def from_result(cls, result, groupby):
        if not isinstance(result, pd.Series):
            raise TypeError("requires pandas Series")

        # Series.groupby is hard-coded to produce a SeriesGroupBy,
        # but it's signature is very large, so use inspect to bind on it.
        sig = inspect.signature(result.groupby)
        bound = sig.bind(by = result.index)
        
        orig_grouper = groupby._orig_grouper if isinstance(groupby, cls) else groupby.grouper
        orig_obj = groupby._orig_obj if isinstance(groupby, cls) else groupby.obj
        should_cast = False
        
        return cls(
            result,
            *bound.args, **bound.kwargs,
            orig_grouper = orig_grouper,
            orig_obj = orig_obj,
            should_cast = should_cast
            )

def broadcast_agg_result(grouper, result, obj, cast = False):
    """
    fast version of transform, only applicable to
    builtin/cythonizable functions
    """
    ids, _, ngroup = grouper.group_info
    out = algorithms.take_1d(result._values, ids)
    
    # TODO: consequence of skipping this step? A cast is already done
    # once when aggregating....
    if cast:
        out = try_cast(out, obj)
    return pd.Series(out, index=obj.index, name=obj.name)


# Utils =======================================================================

def all_isinstance(cls, *args):
    return all(isinstance(x, cls) for x in args)

def _regroup(res, groupby):
    if isinstance(groupby, GroupByAgg):
        # need to manually a constructor, since Series classes are hardcoded
        # all over the pandas library :/ :/ :/
        return groupby.from_result(res, groupby)
    elif isinstance(groupby, SeriesGroupBy):
        return res.groupby(groupby.grouper)
    
    raise ValueError("Unknown group by class: %s"% type(groupby))


# Broadcasting Groupby elements -----------------------------------------------

def grouper_match(grp1, grp2):
    if not isinstance(grp2, SeriesGroupBy):
        raise TypeError("grp2 must be a SeriesGroupBy")

    if grp1._orig_grouper is not grp2.grouper:
        raise ValueError("groups must match")

    return grp1._broadcast_agg_result(), grp2.obj, grp2
    

def broadcast_group_elements(x, y):
    """Returns 3-tuple of same-length x and y data, plus a reference group by object.

    Note:
        * Raises error if x and y are not compatible group by objects.
        * Will broadcast a GroupByAgg, to ensure same length as other data.
    """
    if all_isinstance(GroupByAgg, x, y) and x._orig_grouper is y._orig_grouper:
        return x.obj, y.obj, x
    
    elif isinstance(x, GroupByAgg):
        return grouper_match(x, y)
    
    elif isinstance(y, GroupByAgg):
        res_y, res_x, grp = grouper_match(y, x)
        return res_x, res_y, grp
    
    elif all_isinstance(SeriesGroupBy, x, y) and x.grouper is y.grouper:
        return x.obj, y.obj, x

    raise ValueError("need groupby objects with matching groupers")




import pytest
import pandas as pd
from numpy.testing import assert_equal
from pandas.testing import assert_series_equal, assert_index_equal
from siuba.dply.forcats import fct_recode, fct_collapse

@pytest.fixture
def series1():
    yield pd.Series(["a", "b", "c", "d"])

@pytest.fixture
def cat1():
    yield pd.Categorical(["a", "b", "c", "d"])

# Need to ensure all functions...
#  - 1. can take a series or array
#  - 2. handle a symbolic or call
#  - 3. handle names with spaces

# just a little shorter to write...
def factor(values, categories, ordered = False):
    return pd.Categorical(values, categories, ordered)

def assert_cat_equal(a, b):
    assert isinstance(a, pd.Categorical)
    assert isinstance(b, pd.Categorical)

    assert_index_equal(a.categories, b.categories)
    assert_equal(a.codes, b.codes)
    assert a.ordered == b.ordered

# fct_reorder -----------------------------------------------------------------

@pytest.mark.skip("TODO")
def test_forcats_fct_reorder(cat1):
    pass

# fct_recode ------------------------------------------------------------------

def test_forcats_fct_recode_rename(cat1):
    """Test a single rename"""
    out1 = fct_recode(cat1, x="b")
    out2 = pd.Categorical(["a", "x", "c", "d"], ["a", "x", "c", "d"])
    assert_cat_equal(out1, out2)

@pytest.mark.skip("TODO")
def test_forcats_fct_recode_2_same_name(cat1):
    """Test the case where new categories are defined in kwargs _and_ recat"""
    pytest.raises(ValueError, fct_recode, cat1, recat={"x": "a"}, x="b")


@pytest.mark.skip("TODO")
def test_forcats_fct_recode_dict_and_kwargs(cat1):
    pass

@pytest.mark.skip("TODO")
def test_forcats_fct_recode_remove(cat1):
    """
    # If you name the level NULL it will be removed
    fct_recode(x, NULL = "apple", fruit = "banana")
    """
    pass

@pytest.mark.skip("TODO")
def test_forcats_fct_recode_warn(cat1):
    """
    # If you make a mistake you'll get a warning
    fct_recode(x, fruit = "apple", fruit = "bananana")
    """
    pass

# fct_collapse ----------------------------------------------------------------

def test_forcats_fct_collapse(cat1):
    mapping1 = {
        "x": ["b", "d"],
        "y": "a",
    }
    out1 = fct_collapse(cat1, mapping1)
    out2 = pd.Categorical(["y", "x", "c", "x"], ["y", "x", "c"])

    assert_cat_equal(out1, out2)

# fct_lump --------------------------------------------------------------------

@pytest.mark.skip("TODO")
def test_forcats_fct_lump(cat1):
    pass

# fct_rev ---------------------------------------------------------------------

@pytest.mark.skip("TODO")
def test_forcats_fct_rev(cat1):
    pass



import pandas as pd
import numpy as np
from functools import singledispatch
from siuba.siu import symbolic_dispatch
from pandas.core.groupby import SeriesGroupBy, GroupBy
from pandas.core.frame import NDFrame
from pandas import Series

from siuba.experimental.pd_groups.groupby import GroupByAgg, _regroup
from siuba.experimental.pd_groups.translate import method_agg_op

__ALL__ = [
        "cumall", "cumany", "cummean", 
        "desc",
        "dense_rank", "percent_rank", "min_rank", "cume_dist",
        "row_number",
        "ntile",
        "between",
        "coalesce",
        "lead", "lag",
        "n", "n_distinct",
        "na_if",
        "near",
        "nth", "first", "last"
        ]

# Utils =======================================================================

def _expand_bool(x, f):
    return x.expanding().apply(f, raw = True).astype(bool)

def alias_series_agg(name):
    method = method_agg_op(name, is_property = False, accessor = False)

    def decorator(dispatcher):
        dispatcher.register(SeriesGroupBy, method)
        return dispatcher

    return decorator


# Single dispatch functions ===================================================

# cumall ----------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def cumall(x):
    """Return a same-length array. For each entry, indicates whether that entry and all previous are True-like.

    Example:
        >>> cumall(pd.Series([True, False, False]))
        0     True
        1    False
        2    False
        dtype: bool

    """
    return _expand_bool(x, np.all)


# cumany ----------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def cumany(x):
    """Return a same-length array. For each entry, indicates whether that entry or any previous are True-like.

    Example:
        >>> cumany(pd.Series([False, True, False]))
        0    False
        1     True
        2     True
        dtype: bool

    """
    return _expand_bool(x, np.any)


# cummean ---------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def cummean(x):
    """Return a same-length array, containing the cumulative mean."""
    return x.expanding().mean()


@cummean.register(SeriesGroupBy)
def _cummean_grouped(x):
    grouper = x.grouper
    n_entries = x.obj.notna().groupby(grouper).cumsum()

    res = x.cumsum() / n_entries

    return res.groupby(grouper)


# desc ------------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def desc(x):
    """Return array sorted in descending order."""
    return x.sort_values(ascending = False).reset_index(drop = True)


# dense_rank ------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def dense_rank(x, na_option = "keep"):
    """Return the dense rank.
    
    This method of ranking returns values ranging from 1 to the number of unique entries.
    Ties are all given the same ranking.

    Example:

        >>> dense_rank(pd.Series([1,3,3,5]))
        0    1.0
        1    2.0
        2    2.0
        3    3.0
        dtype: float64


    """
    return x.rank(method = "dense", na_option = na_option)


# percent_rank ----------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def percent_rank(x, na_option = "keep"):
    """Return the percent rank.

    Note:
        Uses minimum rank, and reports the proportion of unique ranks each entry is greater than.

    Examples:
        >>> percent_rank(pd.Series([1, 2, 3]))
        0    0.0
        1    0.5
        2    1.0
        dtype: float64

        >>> percent_rank(pd.Series([1, 2, 2]))
        0    0.0
        1    0.5
        2    0.5
        dtype: float64

        >>> percent_rank(pd.Series([1]))
        0   NaN
        dtype: float64


    """
    return (min_rank(x) - 1) / (x.count() - 1)


# min_rank --------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def min_rank(x, na_option = "keep"):
    """Return the min rank. See pd.Series.rank with method="min" for details.

    """
    return x.rank(method = "min", na_option = na_option)


# cume_dist -------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def cume_dist(x, na_option = "keep"):
    """Return the cumulative distribution corresponding to each value in x.

    This reflects the proportion of values that are less than or equal to each value.

    """
    return x.rank(method = "max", na_option = na_option) / x.count()


# row_number ------------------------------------------------------------------

@symbolic_dispatch(cls = NDFrame)
def row_number(x):
    """Return the row number (position) for each value in x, beginning with 1.

    Example:
        >>> ser = pd.Series([7,8])
        >>> row_number(ser)
        0    1
        1    2
        dtype: int64

        >>> row_number(pd.DataFrame({'a': ser}))
        0    1
        1    2
        dtype: int64

        >>> row_number(pd.Series([7,8], index = [3, 4]))
        3    1
        4    2
        dtype: int64


    """
    if isinstance(x, pd.DataFrame):
        n = x.shape[0]
    else:
        n = len(x)
    
    arr = np.arange(1, n + 1)

    # could use single dispatch, but for now ensure output data type matches input
    if isinstance(x, pd.Series):
        return x._constructor(arr, x.index, fastpath = True)

    return pd.Series(arr, x.index, fastpath = True)


@row_number.register(GroupBy)
def _row_number_grouped(g: GroupBy) -> GroupBy:
    out = np.ones(len(g.obj), dtype = int)

    indices = g.grouper.indices
    for g_key, inds in indices.items():
        out[inds] = np.arange(1, len(inds) + 1, dtype = int)
    
    return _regroup(out, g)


# ntile -----------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def ntile(x, n):
    """TODO: Not Implemented"""
    raise NotImplementedError("ntile not implemented")


# between ---------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def between(x, left, right, default = False):
    """Return whether a value is between left and right (including either side).

    Example:
        >>> between(pd.Series([1,2,3]), 0, 2)
        0     True
        1     True
        2    False
        dtype: bool

    Note:
        This is a thin wrapper around pd.Series.between(left, right)
    
    """
    # note: NA -> False, in tidyverse NA -> NA
    if default is not False:
        raise TypeError("between function must use default = False for pandas Series")

    return x.between(left, right)
    

# coalesce --------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def coalesce(x, *args):
    """Returns a copy of x, with NaN values filled in from *args. Ignores indexes.

    Arguments:
        x: a pandas Series object
        *args: other Series that are the same length as x, or a scalar

    Examples:
        >>> x = pd.Series([1., None, None])
        >>> abc = pd.Series(['a', 'b', None])
        >>> xyz = pd.Series(['x', 'y', 'z'])
        >>> coalesce(x, abc)
        0       1
        1       b
        2    None
        dtype: object

        >>> coalesce(x, abc, xyz)
        0    1
        1    b
        2    z
        dtype: object
        
    """

    crnt = x.reset_index(drop = True)

    for other in args:
        if isinstance(other, pd.Series):
            other = other.reset_index(drop = True)

        crnt = crnt.where(crnt.notna(), other)

    crnt.index = x.index
    return crnt


# lead ------------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def lead(x, n = 1, default = None):
    """Return an array with each value replaced by the next (or further forward) value in the array.

    Arguments:
        x: a pandas Series object
        n: number of next values forward to replace each value with
        default: what to replace the n final values of the array with

    Example:
        >>> lead(pd.Series([1,2,3]), n=1)
        0    2.0
        1    3.0
        2    NaN
        dtype: float64

        >>> lead(pd.Series([1,2,3]), n=1, default = 99)
        0     2
        1     3
        2    99
        dtype: int64

    """
    res = x.shift(-1*n, fill_value = default)

    return res


@lead.register(SeriesGroupBy)
def _lead_grouped(x, n = 1, default = None):
    res = x.shift(-1*n, fill_value = default)

    return _regroup(res, x)


# lag -------------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def lag(x, n = 1, default = None):
    """Return an array with each value replaced by the previous (or further backward) value in the array.

    Arguments:
        x: a pandas Series object
        n: number of next values backward to replace each value with
        default: what to replace the n final values of the array with

    Example:
        >>> lag(pd.Series([1,2,3]), n=1)
        0    NaN
        1    1.0
        2    2.0
        dtype: float64

        >>> lag(pd.Series([1,2,3]), n=1, default = 99)
        0    99.0
        1     1.0
        2     2.0
        dtype: float64


    """
    res = x.shift(n)

    if default is not None:
        res.iloc[:n] = default

    return res


@lag.register(SeriesGroupBy)
def _lag_grouped(x, n = 1, default = None):
    res = x.shift(n, fill_value = default)

    return _regroup(res, x)

# n ---------------------------------------------------------------------------

@symbolic_dispatch(cls = NDFrame)
def n(x):
    """Return the total number of elements in the array (or rows in a DataFrame).

    Example:
        >>> ser = pd.Series([1,2,3])
        >>> n(ser)
        3

        >>> df = pd.DataFrame({'x': ser})
        >>> n(df)
        3

    """
    if isinstance(x, pd.DataFrame):
        return x.shape[0]

    return len(x)


@n.register(GroupBy)
def _n_grouped(x: GroupBy) -> GroupByAgg:
    return GroupByAgg.from_result(x.size(), x)


# n_distinct ------------------------------------------------------------------

@alias_series_agg('nunique')
@symbolic_dispatch(cls = Series)
def n_distinct(x):
    """Return the total number of distinct (i.e. unique) elements in an array.
    
    Example:
        >>> n_distinct(pd.Series([1,1,2,2]))
        2

    """
    return x.nunique()


# na_if -----------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def na_if(x, y):
    """Return a array like x, but with values in y replaced by NAs.
    
    Examples:
        >>> na_if(pd.Series([1,2,3]), [1,3])
        0    NaN
        1    2.0
        2    NaN
        dtype: float64
        
    """
    y = [y] if not np.ndim(y) else y

    tmp_x = x.copy(deep = True)
    tmp_x[x.isin(y)] = np.nan

    return tmp_x


# near ------------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def near(x):
    """TODO: Not Implemented"""
    raise NotImplementedError("near not implemented") 


# nth -------------------------------------------------------------------------

@symbolic_dispatch(cls = Series)
def nth(x, n, order_by = None, default = None):
    """Return the nth entry of x. Similar to x[n].

    Note:
        first(x) and last(x) are nth(x, 0) and nth(x, -1).

    Arguments:
        x: series to get entry from.
        n: position of entry to get from x (0 indicates first entry).
        order_by: optional Series used to reorder x.
        default: (not implemented) value to return if no entry at n.

    Examples:
        >>> ser = pd.Series(['a', 'b', 'c'])
        >>> nth(ser, 1)
        'b'

        >>> sorter = pd.Series([1, 2, 0])
        >>> nth(ser, 1, order_by = sorter)
        'a'

        >>> nth(ser, 0), nth(ser, -1)
        ('a', 'c')

        >>> first(ser), last(ser)
        ('a', 'c')

    """

    if default is not None:
        raise NotImplementedError("default argument not implemented") 

    # check indexing is in range, handles positive and negative cases.
    # TODO: is returning None the correct behavior for an empty Series?
    if n >= len(x) or abs(n) > len(x):
        return default

    if order_by is None:
        return x.iloc[n]

    # case where order_by is specified and n in range ----
    # TODO: ensure order_by is arraylike
    if not isinstance(order_by, pd.Series):
        raise NotImplementedError(
                "order_by argument is type %s, but currently only"
                "implemented for Series" % type(order_by)
                )

    if len(x) != len(order_by):
        raise ValueError("x and order_by arguments must be same length")

    order_indx = order_by.reset_index(drop = True).sort_values().index
    return x.iloc[order_indx[n]]


# first and last ----

from functools import wraps

_copy_nth_docs = wraps(nth, assigned = ('__doc__',))

@_copy_nth_docs
def first(x, order_by = None, default = None):
    return nth(x, 0, order_by, default)


@_copy_nth_docs
def last(x, order_by = None, default = None):
    return nth(x, -1, order_by, default)



from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
from importlib.util import find_spec
import importlib
from types import ModuleType
import sys

from functools import wraps
import pdb
import readline




class CallFinder(MetaPathFinder):
    def __init__(self, f):
        self.enabled = True
        self.f = f

    def find_spec(self, fullname, path, target = None):
        if not self.enabled:
            return None

        pkg, *subpkgs = fullname.split('.')
        if fullname.startswith("siuba.meta_hook") and len(subpkgs) > 1:
        #if pkg == 'meta_hook' and len(subpkgs):
            self.enabled = False
            spec = find_spec(".".join(subpkgs[1:]))
            self.enabled = True
            #spec.loader = CallLoader(self.f, spec.loader, spec)
            return ModuleSpec(fullname, CallLoader(self.f, spec))
        elif pkg == "meta_hook":
            pass

    #def invalidate_caches(self):
    #    pass

class CallLoader(Loader):
    def __init__(self, f, spec):
        self.f = f
        self.orig_spec = spec

    def create_module(self, spec):
        #self.orig_module = self.orig_spec.loader.create_module(self.orig_spec)
        self.orig_module = importlib.import_module(self.orig_spec.name)        
        #self.orig_module = self.orig_loader.create_module(spec)
        #if self.orig_module is None:
        #    self.orig_module = ModuleType(spec.name)

        #return self.orig_module
        return None


    def exec_module(self, module):
        #self.orig_loader.exec_module(self.orig_module)

        #for k,v in self.orig_module.__dict__.items():
        for k,v in self.orig_module.__dict__.items():
            if k.startswith('_'):
                module.__dict__[k] = v
            else:
                module.__dict__[k] = self.f(v)


from .siu import Symbolic, Call

def lazy_func(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return Symbolic(source = f)(*args, **kwargs)

    return wrapper


if sys.meta_path[0].__class__.__name__ == "CallFinder":
    sys.meta_path[0] = CallFinder(lazy_func)
else:
    sys.meta_path = [CallFinder(lazy_func)] + sys.meta_path

__path__ = ''




import pandas as pd
import pkg_resources

_fname = pkg_resources.resource_filename("siuba.data", "mtcars.csv")

mtcars = pd.read_csv(_fname)
mtcars.__doc__ = """
mtcars data.

Source: Henderson and Velleman (1981), Building multiple regression models interactively. Biometrics, 37, 391–411.

--- Original DataFrame docs below ---
""" + mtcars.__doc__



import importlib

def get_dialect_funcs(name):
    #dialect = engine.dialect.name
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    return mod.funcs

def get_sql_classes(name):
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    win_name = name.title() + "Column"
    agg_name = name.title() + "ColumnAgg"
    return {
            'window': getattr(mod, win_name),
            'aggregate': getattr(mod, agg_name)
            }


def mock_sqlalchemy_engine(dialect):
    """
    Create a sqlalchemy.engine.Engine without it connecting to a database.

    Examples
    --------

    ::
        from siuba.sql import LazyTbl
        from siuba import _, mutate, show_query

        engine = mock_sqlalchemy_engine('postgresql')
        tbl = LazyTbl(engine, 'some_table', ['x'])

        query = mutate(tbl, y = _.x + _.x)
        show_query(query)

    """
    from sqlalchemy.engine import Engine
    from sqlalchemy.dialects import registry

    dialect_cls = registry.load('postgresql')   
    return Engine(None, dialect_cls(), '')  




from sqlalchemy import create_engine, types
from siuba.sql import LazyTbl, collect
from siuba.dply.verbs import ungroup
from pandas.testing import assert_frame_equal
import pandas as pd
import os
import numpy as np

def data_frame(**kwargs):
    fixed = {k: [v] if not np.ndim(v) else v for k,v in kwargs.items()}
    return pd.DataFrame(fixed)

BACKEND_CONFIG = {
        "postgresql": {
            "dialect": "postgresql",
            "dbname": ["SB_TEST_PGDATABASE", "postgres"],
            "port": ["SB_TEST_PGPORT", "5433"],
            "user": ["SB_TEST_PGUSER", "postgres"],
            "password": ["SB_TEST_PGPASSWORD", ""],
            "host": ["SB_TEST_PGHOST", "localhost"],
            },
        "sqlite": {
            "dialect": "sqlite",
            "dbname": ":memory:",
            "port": "0",
            "user": "",
            "password": "",
            "host": ""
            }
        }

class Backend:
    def __init__(self, name):
        self.name = name

    def dispose(self):
        pass

    def load_df(self, df = None, **kwargs):
        if df is None and kwargs:
            df = pd.DataFrame(kwargs)
        elif df is not None and kwargs:
            raise ValueError("Cannot pass kwargs, and a DataFrame")

        return df

    def load_cached_df(self, df):
        return df

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, repr(self.name))

class PandasBackend(Backend):
    pass

class SqlBackend(Backend):
    table_name_indx = 0
    sa_conn_fmt = "{dialect}://{user}:{password}@{host}:{port}/{dbname}"

    def __init__(self, name):
        cnfg = BACKEND_CONFIG[name]
        params = {k: os.environ.get(*v) if isinstance(v, (list)) else v for k,v in cnfg.items()}

        self.name = name
        self.engine = create_engine(self.sa_conn_fmt.format(**params))
        self.cache = {}

    def dispose(self):
        self.engine.dispose()

    @classmethod
    def unique_table_name(cls):
        cls.table_name_indx += 1
        return "siuba_{0:03d}".format(cls.table_name_indx)

    def load_df(self, df = None, **kwargs):
        df = super().load_df(df, **kwargs)
        return copy_to_sql(df, self.unique_table_name(), self.engine)

    def load_cached_df(self, df):
        import hashlib
        from pandas import util
        hash_arr = util.hash_pandas_object(df, index=True).values
        hashed = hashlib.sha256(hash_arr).hexdigest()

        if hashed in self.cache:
            return self.cache[hashed]
        
        res = self.cache[hashed] = self.load_df(df)

        return res


def robust_multiple_sort(df, by):
    """Sort a DataFrame on multiple columns, slower but more reliable than df.sort_values

    Note: pandas errors when you sort by multiple columns, and one has unhashable objects.
          however, it can "sort" a single column with unhashable objects.

          e.g. df.sort_values(by = ['a', 'b']) may cause an error

    This implementation chains sort_values on single columns. In this case,
    pandas sorts a list based on its first entry ¯\_(ツ)_/¯.
    """

    from functools import reduce

    out = reduce(lambda data, col: data.sort_values(col), by, df)

    return out.reset_index(drop = True)

def assert_frame_sort_equal(a, b, **kwargs):
    """Tests that DataFrames are equal, even if rows are in different order"""
    df_a = ungroup(a)
    df_b = ungroup(b)
    sorted_a = robust_multiple_sort(df_a, list(df_a.columns)).reset_index(drop = True)
    sorted_b = robust_multiple_sort(df_b, list(df_b.columns)).reset_index(drop = True)

    assert_frame_equal(sorted_a, sorted_b, **kwargs)

def assert_equal_query(tbl, lazy_query, target, **kwargs):
    out = collect(lazy_query(tbl))

    if isinstance(tbl, pd.DataFrame):
        df_a = ungroup(out).reset_index(drop = True)
        df_b = ungroup(target).reset_index(drop = True)
        assert_frame_equal(df_a, df_b, **kwargs)
    else:
        assert_frame_sort_equal(out, target, **kwargs)


PREFIX_TO_TYPE = {
        # for datetime, need to convert to pandas datetime column
        #"dt": types.DateTime,
        "int": types.Integer,
        "float": types.Float,
        "str": types.String
        }

def auto_types(df):
    dtype = {}
    for k in df.columns:
        pref, *_ = k.split('_')
        if pref in PREFIX_TO_TYPE:
            dtype[k] = PREFIX_TO_TYPE[pref]
    return dtype


def copy_to_sql(df, name, engine):
    if isinstance(engine, str):
        engine = create_engine(engine)

    df.to_sql(name, engine, dtype = auto_types(df), index = False, if_exists = "replace")
    return LazyTbl(engine, name)


from functools import wraps
import pytest
        
def backend_notimpl(*names):
    def outer(f):
        @wraps(f)
        def wrapper(backend, *args, **kwargs):
            if backend.name in names:
                with pytest.raises(NotImplementedError):
                    f(backend, *args, **kwargs)
                pytest.xfail("Not implemented!")
            else:
                return f(backend, *args, **kwargs)
        return wrapper
    return outer

def backend_sql(msg):
    # allow decorating without an extra call
    if callable(msg):
        return backend_sql(None)(msg)

    def outer(f):
        @wraps(f)
        def wrapper(backend, *args, **kwargs):
            if not isinstance(backend, SqlBackend):
                pytest.skip(msg)
            else:
                return f(backend, *args, **kwargs)
        return wrapper
    return outer

def backend_pandas(msg):
    # allow decorating without an extra call
    if callable(msg):
        return backend_pandas(None)(msg)

    def outer(f):
        @wraps(f)
        def wrapper(backend, *args, **kwargs):
            if not isinstance(backend, PandasBackend):
                pytest.skip(msg)
            else:
                return f(backend, *args, **kwargs)
        return wrapper
    return outer



import pytest
from .helpers import assert_equal_query, PandasBackend, SqlBackend, data_frame

def pytest_addoption(parser):
    parser.addoption(
            "--dbs", action="store", default="sqlite", help="databases tested against (comma separated)"
            )

params_backend = [
    pytest.param(lambda: SqlBackend("postgresql"), id = "postgresql", marks=pytest.mark.postgresql),
    pytest.param(lambda: SqlBackend("sqlite"), id = "sqlite", marks=pytest.mark.sqlite),
    pytest.param(lambda: PandasBackend("pandas"), id = "pandas", marks=pytest.mark.pandas)
    ]

@pytest.fixture(params = params_backend, scope = "session")
def backend(request):
    return request.param()

@pytest.fixture
def skip_backend(request, backend):
    if request.node.get_closest_marker('skip_backend'):
        mark_args = request.node.get_closest_marker('skip_backend').args
        if backend.name in mark_args:
            pytest.skip('skipped on backend: {}'.format(backend.name)) 




from .groupby import DataFrameGroupBy, GroupByAgg, SeriesGroupBy, broadcast_group_elements, _regroup
import pandas as pd


# utilities -------------------------------------------------------------------

def _validate_data_args(x, **kwargs):
    if not isinstance(x, SeriesGroupBy):
        raise TypeError("First data argument must be a grouped Series object")

    for name, other_data in kwargs.items():
        if isinstance(other_data, pd.Series):
            raise TypeError("{} may not be a Series.".format(name))


def _apply_grouped_method(ser, name, is_property, accessor, args, kwargs):
    if accessor:
        method = getattr(getattr(ser, accessor), name)
    else:
        method = getattr(ser, name)

    res = method(*args, **kwargs) if not is_property else method

    return res


def _maybe_broadcast(x, y):
    """Same as broadcast_group_elements, but y doesn't have to be SeriesGroupBy

    This is important when y is a literal (e.g. 1), since we don't want to raise
    an error, or broadcast 1 to the length of x. Rather, we want to keep the literal,
    and let the pandas series handle it in the operation.

    """
    if isinstance(y, SeriesGroupBy):
        left, right, groupby = broadcast_group_elements(x, y)
    else:
        left, right, groupby = x.obj, y, x

    return left, right, groupby


# Translations ----------------------------------------------------------------

def not_implemented(name, is_property, accessor):
    return NotImplementedError


def method_agg_op(name, is_property, accessor):
    def f(__ser, *args, **kwargs):
        _validate_data_args(__ser)
        res = _apply_grouped_method(__ser, name, is_property, accessor, args, kwargs)

        return GroupByAgg.from_result(res, __ser)

    f.__name__ = f.__qualname__ = name
    return f


def method_el_op(name, is_property, accessor):
    def f(__ser, *args, **kwargs):
        _validate_data_args(__ser)
        res = _apply_grouped_method(__ser.obj, name, is_property, accessor, args, kwargs)

        return _regroup(res, __ser)

    f.__name__ = f.__qualname__ = name
    return f


def method_el_op2(name, is_property, accessor):
    def f(x, y):
        _validate_data_args(x, y = y)
        left, right, groupby = _maybe_broadcast(x, y)
        
        op_function = getattr(left, name)

        res = op_function(right)
        return _regroup(res, groupby)

    f.__name__ = f.__qualname__ = name
    return f


def method_win_op(name, is_property, accessor):
    def f(__ser, *args, **kwargs):
        _validate_data_args(__ser)
        res = _apply_grouped_method(__ser, name, is_property, accessor, args, kwargs)

        return _regroup(res, __ser)

    f.__name__ = f.__qualname__ = name
    return f


GROUP_METHODS = { 
        ("Elwise", 1): method_el_op,
        ("Elwise", 2): method_el_op2,
        ("Agg", 1): method_agg_op,
        ("Window", 1): method_win_op,
        ("Window", 2): method_win_op,
        ("Singleton", 1): not_implemented
        }






from siuba import _, group_by, mutate, filter, summarize
from siuba.sql import sql_raw
import sqlalchemy.exc

import pytest

from .helpers import assert_equal_query, data_frame, backend_sql

DATA = data_frame(x = ['a','a'], y = [1,2])

@pytest.fixture(scope = "module")
def df(backend):
    return backend.load_df(DATA)


@backend_sql
def test_raw_sql_mutate(backend, df):
    assert_equal_query(
            df,
            mutate(z = sql_raw("y + 1")),
            DATA.assign(z = lambda d: d.y + 1)
            )


@backend_sql
def test_raw_sql_mutate_grouped(backend, df):
    assert_equal_query(
            df,
            group_by("x") >> mutate(z = sql_raw("y + 1")),
            DATA.assign(z = lambda d: d.y + 1)
            )


@backend_sql
def test_raw_sql_mutate_refer_previous_raise_dberror(backend, df):
    # Note: unlikely will be able to support this case. Normally we analyze
    # the expression to know whether we need to create a subquery.
    with pytest.raises(sqlalchemy.exc.DatabaseError):
        assert_equal_query(
                df,
                group_by("x") >> mutate(z1 = sql_raw("y + 1"), z2 = sql_raw("z1 + 1")),
                DATA.assign(z1 = lambda d: d.y + 1, z2 = lambda d: d.z1 + 1)
                )

@backend_sql
def test_raw_sql_filter(backend, df):
    assert_equal_query(
            df,
            filter(sql_raw("y = 1")),
            data_frame(x = ['a'], y = [1])
            )

@backend_sql
def test_raw_sql_summarize(backend, df):
    assert_equal_query(
            df, 
            summarize(z = sql_raw("SUM(y)")) >> mutate(z = _.z.astype(int)),
            data_frame(z = [3])
            )




"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-select.R
"""
    
from siuba import _, mutate, select, group_by, rename

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql
from string import ascii_lowercase 

DATA = data_frame(a = 1, b = 2, c = 3)

@pytest.fixture(scope = "module")
def dfs(backend):
    return backend.load_df(DATA)

@pytest.mark.parametrize("query, output", [
    ( select(_.c), data_frame(c = 3) ),
    ( select(_.b == _.c), data_frame(b = 3) ),
    ( select(_["a":"c"]), data_frame(a = 1, b = 2, c = 3) ),
    ( select(_[_.a:_.c]), data_frame(a = 1, b = 2, c = 3) ),
    ( select(_.a, _.b) >> select(_.b), data_frame(b = 2) ),
    ( mutate(a = _.b + _.c) >> select(_.a), data_frame(a = 5) ),
    pytest.param( group_by(_.a) >> select(_.b), data_frame(b = 2, a = 1), marks = pytest.mark.xfail),
    ])
def test_select_siu(dfs, query, output):
    assert_equal_query(dfs, query, output)


@pytest.mark.skip("TODO: #63")
def test_select_kwargs(dfs):
    assert_equal_query(dfs, select(x = _.a), data_frame(x = 1))


# Rename ----------------------------------------------------------------------

@pytest.mark.parametrize("query, output", [
    ( rename(A = _.a), data_frame(A = 1, b = 2, c = 3) ),
    ( rename(A = "a"), data_frame(A = 1, b = 2, c = 3) ),
    ( rename(A = _.a, B = _.c), data_frame(A = 1, b = 2, B = 3) ),
    ( rename(A = "a", B = "c"), data_frame(A = 1, b = 2, B = 3) )
    ])
def test_rename_siu(dfs, query, output):
    assert_equal_query(dfs, query, output)


@backend_sql("TODO: pandas - grouped df rename")
@pytest.mark.parametrize("query, output", [
    ( group_by(_.a) >> rename(z = _.a), data_frame(z = 1, b = 2, c = 3) ),
    ( group_by(_.a) >> rename(z = "a"), data_frame(z = 1, b = 2, c = 3) )
    ])
def test_grouped_rename_siu(backend, dfs, query, output):
    assert_equal_query(dfs, query, output)




"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-mutate.R
"""
    
from siuba import _, group_by, summarize, count
import pandas as pd

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql

DATA = data_frame(x = [1,2,3,4], g = ['a', 'a', 'b', 'b'])
DATA2 = data_frame(x = [1,2,3,4], g = ['a', 'a', 'b', 'b'], h = ['c', 'c', 'd', 'd'])

@pytest.fixture(scope = "module")
def df(backend):
    return backend.load_df(DATA)

@pytest.fixture(scope = "module")
def df2(backend):
    return backend.load_df(DATA2)


@pytest.mark.parametrize("query, output", [
    (count(_.g), data_frame(g = ['a', 'b'], n = [2, 2])),
    (count("g"), data_frame(g = ['a', 'b'], n = [2, 2])),
    (count("x", "g"), DATA.assign(n = 1)),
    (count(_.x, "g"), DATA.assign(n = 1))
    ])
def test_basic_count(df, query, output):
    assert_equal_query(df, query, output)


@pytest.mark.skip("TODO: sql fix unnamed expression labels in count (#69)")
def test_count_with_expression(df):
    assert_equal_query(
            df,
            count(_.x - _.x),
            pd.DataFrame({"x - x": [0], "n": [4]})
            )


@pytest.mark.skip("TODO: sql support kwargs in count (#68)")
def test_count_with_kwarg_expression(df):
    assert_equal_query(
            df,
            count(y = _.x - _.x),
            pd.DataFrame({"y": [0], "n": [4]})
            )

@backend_notimpl("sqlite", "postgresql")     # see (#104)
def test_count_wt(backend, df):
    assert_equal_query(
            df,
            count(_.g, wt = _.x),
            pd.DataFrame({'g': ['a', 'b'], 'n': [1 + 2, 3 + 4]})
            )

def test_count_no_groups(df):
    # count w/ no groups returns ttl
    assert_equal_query(
            df,
            count(),
            pd.DataFrame({'n': [4]})
            )

@backend_notimpl("sqlite", "postgresql")   # see (#104)
def test_count_no_groups_wt(backend, df):
    assert_equal_query(
            df,
            count(wt = _.x),
            pd.DataFrame({'n': [sum([1,2,3,4])]})
            )


def test_count_on_grouped_df(df2):
    assert_equal_query(
            df2,
            group_by(_.g) >> count(_.h),
            pd.DataFrame({'g': ['a', 'b'], 'h': ['c', 'd'], 'n': [2,2]})
            )





from siuba.sql.utils import get_dialect_funcs
import pytest

@pytest.mark.parametrize('name', [
    'redshift',
    'postgresql',
    'sqlite'
    ])
def test_get_dialect_funcs(name):
    get_dialect_funcs(name)



"""
This module holds default translations from pandas syntax to sql for 3 kinds of operations...

1. scalar - elementwise operations (e.g. array1 + array2)
2. aggregation - operations that result in a single number (e.g. array1.mean())
3. window - operations that do calculations across a window
            (e.g. array1.lag() or array1.expanding().mean())


"""

from sqlalchemy import sql
from sqlalchemy.sql import sqltypes as types, func as fn
from functools import singledispatch
from .verbs import case_when, if_else

# warning for when sql defaults differ from pandas ============================
import warnings


class SiubaSqlRuntimeWarning(UserWarning): pass

def warn_arg_default(func_name, arg_name, arg, correct):
    warnings.warn(
            "\n{func_name} sql translation defaults "
            "{arg_name} to {arg}. To return identical result as pandas, use "
            "{arg_name} = {correct}.\n\n"
            "This warning only displays once per function".format(
                func_name = func_name, arg_name = arg_name, arg = repr(arg), correct = repr(correct)
                ),
            SiubaSqlRuntimeWarning
            )

# Custom dispatching in call trees ============================================
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.base import ImmutableColumnCollection

class SqlColumn(ColumnClause): pass

class SqlColumnAgg(SqlColumn): pass

# Custom over clause handling  ================================================

# TODO: must make these take both tbl, col as args, since hard to find window funcs
def sa_is_window(clause):
    return isinstance(clause, sql.elements.Over) \
            or isinstance(clause, sql.elements.WithinGroup)

def sa_get_over_clauses(clause):
    windows = []
    append_win = lambda col: windows.append(col)

    sql.util.visitors.traverse(clause, {}, {"over": append_win})

    return windows

def sa_modify_window(clause, group_by = None, order_by = None):
    if group_by:
        group_cols = [columns[name] for name in group_by]
        partition_by = sql.elements.ClauseList(*group_cols)
        clone = clause._clone()
        clone.partition_by = partition_by

        return clone

    return clause

from sqlalchemy.sql.elements import Over

class CustomOverClause: pass

class AggOver(Over, CustomOverClause):
    def set_over(self, group_by, order_by = None):
        self.partition_by = group_by
        return self


class RankOver(Over, CustomOverClause): 
    def set_over(self, group_by, order_by = None):
        crnt_partition = getattr(self.partition_by, 'clauses', tuple())
        self.partition_by = sql.elements.ClauseList(*crnt_partition, *group_by.clauses)
        return self


class CumlOver(Over, CustomOverClause):
    def set_over(self, group_by, order_by):
        self.partition_by = group_by
        self.order_by = order_by

        if not len(order_by):
            warnings.warn(
                    "No order by columns explicitly set in window function. SQL engine"
                    "does not guarantee a row ordering. Recommend using an arrange beforehand.",
                    RuntimeWarning
                    )
        return self


# Translator creation funcs ===================================================

# Windows ----

def win_absent(name):
    from typing import Any
    def not_implemented(*args, **kwargs) -> Any:
        raise NotImplementedError("SQL dialect does not support {}.".format(name))
    
    return not_implemented

def win_over(name: str):
    sa_func = getattr(sql.func, name)
    def f(col) -> RankOver:
        return RankOver(sa_func(), order_by = col)

    return f

def win_cumul(name):
    sa_func = getattr(sql.func, name)
    def f(col, *args, **kwargs) -> CumlOver:
        return CumlOver(sa_func(col, *args, **kwargs), rows = (None,0))

    return f

def win_agg(name):
    sa_func = getattr(sql.func, name)
    def f(col, *args, **kwargs) -> AggOver:
        return AggOver(sa_func(col, *args, **kwargs))

    return f

def sql_func_diff(col, periods = 1):
    if periods > 0:
        return CumlOver(col - sql.func.lag(col, periods))
    elif periods < 0:
        return CumlOver(col - sql.func.lead(col, abs(periods)))

    raise ValueError("periods argument to sql diff cannot be 0")


# Ordered and theoretical set aggregates ----

def set_agg(name):
    sa_func = getattr(sql.func, name)
    return lambda col, *args: sa_func(*args).within_group(col)

# Datetime ----

def sql_extract(name):
    return lambda col: sql.func.extract(name, col)

def sql_func_extract_dow_monday(col):
    # make monday = 0 rather than sunday
    monday0 = sql.cast(sql.func.extract('dow', col) + 6, types.Integer) % 7
    # cast to numeric, since that's what extract('dow') returns
    return sql.cast(monday0, types.Numeric)

def sql_is_first_of(name, reference):
    return lambda col: fn.date_trunc(name, col) == fn.date_trunc(reference, col)

def sql_func_last_day_in_period(col, period):
    return fn.date_trunc(period, col) + sql.text("interval '1 %s - 1 day'" % period)

def sql_func_days_in_month(col):
    return fn.extract('day', sql_func_last_day_in_period(col, 'month'))

def sql_is_last_day_of(period):
    def f(col):
        last_day = sql_func_last_day_in_period(col, period)
        return fn.date_trunc('day', col) == last_day

    return f

def sql_func_floor_date(col, unit):
    # see https://www.postgresql.org/docs/9.1/functions-datetime.html#FUNCTIONS-DATETIME-TRUNC
    # valid values: 
    #   microseconds, milliseconds, second, minute, hour,
    #   day, week, month, quarter, year, decade, century, millennium
    # TODO: implement in siuba.dply.lubridate
    return fn.date_trunc(unit, col)


# Strings ----

def sql_str_strip(name):
    
    strip_func = getattr(fn, name)
    def f(col, to_strip = " \t\n\v\f\r"):
        return strip_func(col, to_strip)

    return f

def sql_func_capitalize(col):
    first_char = fn.upper(fn.left(col, 1)) 
    rest = fn.right(col, fn.length(col) - 1)
    return first_char.op('||')(rest)


# Others ----

def sql_agg(name):
    sa_func = getattr(sql.func, name)
    return lambda col: sa_func(col)

def sql_scalar(name):
    sa_func = getattr(sql.func, name)
    return lambda col, *args: sa_func(col, *args)

def sql_colmeth(meth, *outerargs):
    def f(col, *args) -> SqlColumn:
        return getattr(col, meth)(*outerargs, *args)
    return f

def sql_not_impl():
    return NotImplementedError


# Custom implementations ----

def sql_func_astype(col, _type):
    mappings = {
            str: types.Text,
            'str': types.Text,
            int: types.Integer,
            'int': types.Integer,
            float: types.Numeric,
            'float': types.Numeric,
            bool: types.Boolean,
            'bool': types.Boolean
            }
    try:
        sa_type = mappings[_type]
    except KeyError:
        raise ValueError("sql astype currently only supports type objects: str, int, float, bool")
    return sql.cast(col, sa_type)


# Base translations ===========================================================

base_scalar = dict(
        # infix operators -----
        # NOTE: all sqlalchemy.Column operators are used
        # TODO: need way to implement ** operator

        # sqlalchemy.ColumnElement methods ----
        cast = sql_colmeth("cast"),
        between = sql_colmeth("between"),
        isin = sql_colmeth("in_"),

        # bitwise operations -----
        # TODO

        # these are methods on sql.funcs ---- 
        abs = sql_scalar("abs"),
        acos = sql_scalar("acos"),
        asin = sql_scalar("asin"),
        atan = sql_scalar("atan"),
        atan2 = sql_scalar("atan2"),
        cos = sql_scalar("cos"),
        cot = sql_scalar("cot"),
        astype = sql_func_astype,
        # I was lazy here and wrote lambdas directly ---
        # TODO: I think these are postgres specific?
        isna = sql_colmeth("is_", None),
        isnull = sql_colmeth("is_", None),
        notna = lambda col: ~col.is_(None),
        fillna = sql.functions.coalesce,
        # dply.vector funcs ----

        # TODO: move to postgres specific
        # TODO: this is to support a DictCall (e.g. used in case_when)
        dict = dict,
        # TODO: don't use singledispatch to add sql support to case_when
        case_when = case_when,
        if_else = if_else,

        # POSTGRES compatility ------------------------------------------------
        # _special_methods
        __round__ = sql_scalar("round"),

        #
        copy = sql_not_impl(),

        # binary
        add = sql_colmeth('__add__'),
        sub = sql_colmeth('__sub__'),
        #truediv
        #floordiv
        mul = sql_colmeth('__mul__'),
        mod = sql_colmeth('__mod__'),
        #pow = sql_colmeth('__pow__'),
        lt = sql_colmeth('__lt__'),
        gt = sql_colmeth('__gt__'),
        le = sql_colmeth('__le__'),
        ge = sql_colmeth('__ge__'),
        ne = sql_colmeth('__ne__'),
        eq = sql_colmeth('__eq__'),
        #round = sql_scalar("round"),
        radd = sql_colmeth('__radd__'),
        rsub = sql_colmeth('__rsub__'),
        #rtruediv
        #rfloordiv
        rmul = sql_colmeth('__rmul__'),
        rmod = sql_colmeth('__rmod__'),

        # computations ---
        clip = lambda col, lower, upper: fn.least(fn.greatest(col, lower), upper),

        # datetime_properties ---
        date = sql_not_impl(),
        time = sql_not_impl(),
        timetz = sql_not_impl(),
        year = sql_extract('year'),# , sql.cast(col, sql.sqltypes.Date)),
        month = sql_extract('month'),
        day = sql_extract('day'),
        hour = sql_extract('hour'),
        minute = sql_extract('minute'),
        second = sql_extract('second'),
        #microsecond = sql_extract('microsecond'), # TODO: postgres includes seconds
        nanosecond = sql_not_impl(),
        week = sql_extract('week'),
        weekofyear = sql_extract('week'),
        dayofweek = sql_func_extract_dow_monday,
        weekday = sql_func_extract_dow_monday,
        dayofyear = sql_extract('doy'),
        quarter = sql_extract('quarter'),
        is_month_start = sql_is_first_of('day', 'month'),
        is_month_end = sql_is_last_day_of('month'),
        is_quarter_start = sql_is_first_of('day', 'quarter'),
        #is_quarter_end = sql_is_last_day_of('quarter'),
        is_year_start = sql_is_first_of('day', 'year'),
        is_year_end = sql_is_last_day_of('year'),
        is_leap_year = sql_not_impl(),
        daysinmonth = sql_func_days_in_month,
        days_in_month = sql_func_days_in_month,
        tz = sql_not_impl(),
        freq = sql_not_impl(),

        # datetime methods ---
        #to_period = ,
        ## dt.to_pydatetime
        #tz_localize = 
        ## dt.tz_convert
        #normalize = 
        #strftime = 
        #round = 
        #floor = 
        #ceil = 
        #month_name = 
        #day_name =
        # TODO: slotting in a floor_date method, since I can't do my job
        #       or make common SQL queries without it....
        floor_date = sql_func_floor_date,


        # string methdos ---
        
        capitalize = sql_func_capitalize,
        #casefold = ,
        #cat  = ,
        center = sql_not_impl(),
        contains = sql_not_impl(),
#        count = ,
#        # str.decode
#        encode = ,
        endswith = sql_colmeth("endswith"),
#        #extract = 
#        # str.extractall
#        find = ,
#        findall = ,
#        #get
#        #index
#        #join
        len = sql.func.length,
#        ljust = ,
        lower = sql.func.lower,
        # TODO: whitespace options based loosely on builtin string.isspace
        lstrip = sql_str_strip('ltrim'),
#        match = ,
#        # str.normalize
#        pad = ,
#        # str.partition
#        # str.repeat
#        replace = ,
#        rfind = ,
#        # str.rindex
#        rjust = ,
#        # str.rpartition
        rstrip = sql_str_strip('rtrim'),
#        slice = ,
#        slice_replace = ,
#        split = ,
#        rsplit = ,
        startswith = sql_colmeth("startswith"),
        strip = sql_str_strip('trim'),
#        swapcase = ,
        title = sql.func.initcap,
#        # str.translate
        upper = sql.func.upper,
#        wrap = ,
#        # str.zfill
#        isalnum = ,
#        isalpha = ,
#        isdigit = ,
#        isspace = ,
#        islower = ,
#        isupper = ,
#        istitle = ,
#        isnumeric = ,
#        isdecimal = ,
        )

base_agg = dict(
        mean = sql_agg("avg"),
        sum = sql_agg("sum"),
        min = sql_agg("min"),
        max = sql_agg("max"),
        count = sql_agg("count"),
        # TODO: generalize case where doesn't use col
        # need better handeling of vector funcs
        nunique = lambda col: sql.func.count(sql.func.distinct(col)),

        # POSTGRES compatibility ----------------------------------------------
        quantile = set_agg("percentile_cont"),
        )

base_win = dict(
        rank = win_over("rank"),
        #first = win_over2("first"),
        #last = win_over2("last"),
        #nth = win_over3
        #lead = win_over4
        #lag

        # aggregate functions ---
        mean = win_agg("avg"),
        var = win_agg("variance"),
        sum = win_agg("sum"),
        min = win_agg("min"),
        max = win_agg("max"),

        # ordered set funcs ---
        #quantile
        #median

        # counts ----
        count = win_agg("count"),
        #n
        #n_distinct

        # cumulative funcs ---
        #avg("id") OVER (PARTITION BY "email" ORDER BY "id" ROWS UNBOUNDED PRECEDING)
        #cummean = win_agg("
        cumsum = win_cumul("sum"),
        #cummin
        #cummax
        diff = sql_func_diff,

        # POSTGRES compatibility ----------------------------------------------
        # computations
        #prod = lambda col: AggOver(fn.exp(fn.sum(fn.log(col)))),
        std = win_agg("stddev_samp"),
        )

# based on https://github.com/tidyverse/dbplyr/blob/master/R/backend-.R
base_nowin = dict(
        #row_number   = win_absent("ROW_NUMBER"),
        #min_rank     = win_absent("RANK"),
        rank         = win_absent("RANK"),
        dense_rank   = win_absent("DENSE_RANK"),
        percent_rank = win_absent("PERCENT_RANK"),
        cume_dist    = win_absent("CUME_DIST"),
        ntile        = win_absent("NTILE"),
        mean         = win_absent("AVG"),
        sd           = win_absent("SD"),
        var          = win_absent("VAR"),
        cov          = win_absent("COV"),
        cor          = win_absent("COR"),
        sum          = win_absent("SUM"),
        min          = win_absent("MIN"),
        max          = win_absent("MAX"),
        median       = win_absent("PERCENTILE_CONT"),
        quantile    = win_absent("PERCENTILE_CONT"),
        n            = win_absent("N"),
        n_distinct   = win_absent("N_DISTINCT"),
        cummean      = win_absent("MEAN"),
        cumsum       = win_absent("SUM"),
        cummin       = win_absent("MIN"),
        cummax       = win_absent("MAX"),
        nth          = win_absent("NTH_VALUE"),
        first        = win_absent("FIRST_VALUE"),
        last         = win_absent("LAST_VALUE"),
        lead         = win_absent("LEAD"),
        lag          = win_absent("LAG"),
        order_by     = win_absent("ORDER_BY"),
        str_flatten  = win_absent("STR_FLATTEN"),
        count        = win_absent("COUNT")
        )

funcs = dict(scalar = base_scalar, aggregate = base_agg, window = base_win)

# MISC ===========================================================================
# scalar, aggregate, window #no_win, agg_no_win

# local to table
# alias to LazyTbl.no_win.dense_rank...
#          LazyTbl.agg.dense_rank...

from collections.abc import MutableMapping
import itertools

class SqlTranslator(MutableMapping):
    def __init__(self, d, **kwargs):
        self.d = d
        self.kwargs = kwargs

    def __len__(self):
        return len(set(self.d) + set(self.kwargs))

    def __iter__(self):
        old_keys = iter(self.d)
        new_keys = (k for k in self.kwargs if k not in self.d)
        return itertools.chain(old_keys, new_keys)

    def __getitem__(self, x):
        try:
            return self.kwargs[x]
        except KeyError:
            return self.d[x]

    def __setitem__(self, k, v):
        self.d[k] = v

    def __delitem__(self, k):
        del self.d[k]
    



from siuba.siu import Symbolic, strip_symbolic
# TODO: dot, corr, cov
# ordered set aggregate. e.g. mode()
# hypothetical-set aggregate (e.g. rank(a) as if it were in partition(order_by b))

# kinds of windows:
#   * result len n_elements: rank()
#   * result len 1: is_monotonic (lag, diff, and any). ordered set aggregate.
#   * result len input len: percentile_cont([.1, .2]). hypo set aggregate.

_ = Symbolic()

class Result:
    def __init__(self, **kwargs):
        self.options = kwargs

    def to_dict(self):
        return {'type': self.__class__.__name__, **self.options}

class Elwise(Result): pass
class Agg(Result): pass 
class Window(Result): pass
class Singleton(Result): pass
class WontImplement(Result): pass

CATEGORIES_TIME = {
        'time_series', 'datetime_properties', 'datetime_methods', 'period_properties',
        'timedelta_properties', 'timedelta_methods'
        }

CATEGORIES_STRING = {
        'string_methods'
        }


# TODO: test cases
#   * doesn't work in mutate (no_mutate = ['postgresql'])
#   * returns float rather than int (sql_type = 'float')
#   * requires boolean or float input (op = 'bool')
#   * won't be implemented (in some backend) (postgresql = 'not_impl')
#   * isn't implemented now, but will be (postgresql = 'xfail')
funcs = {
    ## ------------------------------------------------------------------------
    # Attributes 
    ## ------------------------------------------------------------------------
    '_special_methods': {
        '__invert__': _.__invert__()         >> Elwise(op = 'bool'),
        '__and__': _.__and__(_)              >> Elwise(op = 'bool'),
        '__or__': _.__or__(_)                >> Elwise(op = 'bool'),
        '__xor__': _.__xor__(_)              >> Elwise(op = 'bool', postgresql = 'xfail'),
        '__neg__': _.__neg__()               >> Elwise(),
        '__pos__': _.__pos__()               >> Elwise(postgresql = 'xfail'),
        '__rand__': _.__rand__(_)            >> Elwise(op = 'bool'),
        '__ror__': _.__ror__(_)              >> Elwise(op = 'bool'),
        '__rxor__': _.__rxor__(_)            >> Elwise(op = 'bool', postgresql = 'xfail'),
        # copied from binary section below
        '__add__': _.__add__(_)               >> Elwise(),
        '__sub__': _.__sub__(_)               >> Elwise(),
        '__truediv__': _.__truediv__(_)       >> Elwise(postgresql = 'xfail'),  # TODO: pg needs cast int to float?
        '__floordiv__': _.__floordiv__(_)     >> Elwise(postgresql = 'xfail'),
        '__mul__': _.__mul__(_)               >> Elwise(),
        '__mod__': _.__mod__(_)               >> Elwise(),
        '__pow__': _.__pow__(_)               >> Elwise(postgresql = 'xfail'),
        '__lt__': _.__lt__(_)                 >> Elwise(),
        '__gt__': _.__gt__(_)                 >> Elwise(),
        '__le__': _.__le__(_)                 >> Elwise(),
        '__ge__': _.__ge__(_)                 >> Elwise(), 
        '__ne__': _.__ne__(_)                 >> Elwise(), 
        '__eq__': _.__eq__(_)                 >> Elwise(), 
        '__div__': _.__div__(_)               >> Elwise(postgresql = 'xfail'),  # TODO: deprecated in python3, not in siu
        '__round__': _.__round__(2)           >> Elwise(postgresql = 'xfail'),  # TODO: pg returns float
        '__radd__': _.__radd__(_)             >> Elwise(), 
        '__rsub__': _.__rsub__(_)             >> Elwise(), 
        '__rmul__': _.__rmul__(_)             >> Elwise(), 
        '__rdiv__': _.__rdiv__(_)             >> Elwise(postgresql = 'xfail'), 
        '__rtruediv__': _.__rtruediv__(_)     >> Elwise(postgresql = 'xfail'),
        '__rfloordiv__': _.__rfloordiv__(_)   >> Elwise(postgresql = 'xfail'),
        '__rmod__': _.__rmod__(_)             >> Elwise(),
        '__rpow__': _.__rpow__(_)             >> Elwise(postgresql = 'xfail'),
        },
    'attributes': {
        # method
        # index
        # array
        # values
        # dtype
        # ftype
        # shape
        # nbytes
        # ndim
        # size                              # TODO: good to support, is a method on grouped series...
        # strides
        # itemsize
        # base
        # T
        # memory_usage
        # hasnans
        # flags
        # empty
        # dtypes
        # ftypes
        # data
        # is_copy
        # name
        # put
        },
    ## ------------------------------------------------------------------------
    # Conversion 
    ## ------------------------------------------------------------------------
    'conversion': {
        'astype': _.astype('str')       >> Elwise(),
        # infer_objects
        'copy': _.copy()              >> Elwise(postgresql = 'not_impl'),
        # bool
        # to_numpy
        # to_period
        # to_timestamp
        # to_list
        # get_values
        # __array__
        },
    ## ------------------------------------------------------------------------
    # Indexing, iteration 
    ## ------------------------------------------------------------------------
    'indexing': {
        # get
        # at
        # iat
        # loc
        # iloc
        # __iter__
        # items
        # iteritems
        # keys
        # pop
        # item
        # xs
        },
    ## ------------------------------------------------------------------------
    # Binary operator functions 
    ## ------------------------------------------------------------------------
    'binary': {
        'add': _.add(_)               >> Elwise(),
        'sub': _.sub(_)               >> Elwise(),
        'truediv': _.truediv(_)       >> Elwise(postgresql = 'xfail'),
        'floordiv': _.floordiv(_)     >> Elwise(postgresql = 'xfail'),
        'mul': _.mul(_)               >> Elwise(),
        'mod': _.mod(_)               >> Elwise(),
        'pow': _.pow(_)               >> Elwise(postgresql = 'xfail'),
        'lt': _.lt(_)                 >> Elwise(),
        'gt': _.gt(_)                 >> Elwise(),
        'le': _.le(_)                 >> Elwise(),
        'ge': _.ge(_)                 >> Elwise(), 
        'ne': _.ne(_)                 >> Elwise(), 
        'eq': _.eq(_)                 >> Elwise(), 
        'div': _.div(_)               >> Elwise(postgresql = 'xfail'), 
        'round': _.round(2)           >> Elwise(postgresql = 'xfail'), 
        'radd': _.radd(_)             >> Elwise(), 
        'rsub': _.rsub(_)             >> Elwise(), 
        'rmul': _.rmul(_)             >> Elwise(), 
        'rdiv': _.rdiv(_)             >> Elwise(postgresql = 'xfail'), 
        'rtruediv': _.rtruediv(_)     >> Elwise(postgresql = 'xfail'),
        'rfloordiv': _.rfloordiv(_)   >> Elwise(postgresql = 'xfail'),
        'rmod': _.rmod(_)             >> Elwise(),
        'rpow': _.rpow(_)             >> Elwise(postgresql = 'xfail'),
        # combine
        # combine_first
        #'product': _.product()        >> Agg(),   # TODO: doesn't exist on GroupedDataFrame
        #'dot': _.dot(_)               >> Agg(),
        },
    ## ------------------------------------------------------------------------
    # Function application, groupby & window 
    ## ------------------------------------------------------------------------
    'function_application': {
        # apply
        # agg
        # aggregate
        # transform
        # map
        # groupby
        # rolling
        # expanding
        # ewm
        # pipe
        },
    ## ------------------------------------------------------------------------
    ## Computations / descriptive stats
    ## ------------------------------------------------------------------------
    'computations': {
        'abs': _.abs()                >> Elwise(),
        'all': _.all()                >> Agg(op = 'bool'),
        'any': _.any()                >> Agg(op = 'bool'),
        #'autocorr': _.autocorr()      >> Window(), # TODO: missing on GDF
        'between': _.between(2, 5)    >> Elwise(),
        'clip': _.clip(2, 5)          >> Elwise(),
        # clip_lower                                # TODO: deprecated
        # clip_upper                                # TODO: deprecated
        #'corr': _.corr(_)             >> Agg(),
        'count': _.count()            >> Agg(),
        #'cov': _.cov(_)               >> Agg(),
        'cummax': _.cummax()          >> Window(postgresql = 'xfail'),
        'cummin': _.cummin()          >> Window(postgresql = 'xfail'),
        'cumprod': _.cumprod()        >> Window(postgresql = 'xfail'),
        'cumsum': _.cumsum()          >> Window(postgresql = 'xfail'),
        # describe
        'diff': _.diff()              >> Window(),
        # factorize
        # 'kurt': _.kurt()              >> Agg(),  # TODO: doesn't exist on GDF
        'mad': _.mad()                >> Agg(postgresql = 'xfail'),
        'max': _.max()                >> Agg(),
        'mean': _.mean()              >> Agg(),
        'median': _.median()          >> Agg(postgresql = 'xfail'),
        'min': _.min()                >> Agg(),
        #'mode': _.mode()              >> Agg(),   # TODO: doesn't exist on GDF, can return > 1 result
        #'nlargest': _.nlargest()      >> Window(),
        #'nsmallest': _.nsmallest()    >> Window(),
        'pct_change': _.pct_change()  >> Window(postgresql = 'xfail'),
        'prod': _.prod()              >> Agg(postgresql = 'xfail'),
        'quantile': _.quantile(.75)      >> Agg(no_mutate = ['postgresql']),
        'rank': _.rank()              >> Window(postgresql = 'xfail'),
        'sem': _.sem()                >> Agg(postgresql = 'xfail'),
        'skew': _.skew()              >> Agg(postgresql = 'xfail'),
        'std': _.std()                >> Agg(),
        'sum': _.sum()                >> Agg(postgresql = 'xfail'), # TODO: pg returns float
        'var': _.var()                >> Agg(),
        #'kurtosis': _.kurtosis()      >> Agg(),  # TODO: doesn't exist on GDF
        # unique
        'nunique': _.nunique()        >> Agg(no_mutate = ['postgresql']),
        #'is_unique': _.is_unique      >> Agg(),  # TODO: all is_... properties not on GDF
        #'is_monotonic': _.is_monotonic >> Agg(),
        #'is_monotonic_increasing': _.is_monotonic_increasing >> Agg(),
        #'is_monotonic_decreasing': _.is_monotonic_decreasing >> Agg(),
        # value_counts
        # compound
        },
    ## ------------------------------------------------------------------------
    # Reindexing / selection / label manipulation 
    ## ------------------------------------------------------------------------
    'reindexing': {
        # align
        # drop
        # droplevel
        # drop_duplicates
        # duplicated
        # equals
        #'first': _.first()            >> Window(),
        # head
        # idxmax
        # idxmin
        'isin': _.isin(tuple([1,2]))  >> Elwise(),
        #'last': _.last()              >> Window(),
        # reindex
        # reindex_like
        # rename
        # rename_axis
        # reset_index
        # sample
        # set_axis
        # take
        # tail
        # truncate
        # where
        # mask
        # add_prefix
        # add_suffix
        # filter
        },
    ## ------------------------------------------------------------------------
    # Missing data handling 
    ## ------------------------------------------------------------------------
    'missing_data': {
        'isna':  _.isna()             >> Elwise(),
        'notna': _.notna()            >> Elwise(),
        # dropna
        'fillna': _.fillna(1)         >> Elwise(),
        # interpolate
        },
    ## ------------------------------------------------------------------------
    # Reshaping, sorting 
    ## ------------------------------------------------------------------------
    'reshaping': {
        # argsort
        # argmin
        # argmax
        # reorder_levels
        # sort_values
        # sort_index
        # swaplevel
        # unstack
        # explode
        # searchsorted
        # ravel
        # repeat
        # squeeze
        # view
        },
    ## ------------------------------------------------------------------------
    # Combining / joining / merging 
    ## ------------------------------------------------------------------------
    'combining': {
        # append
        # replace
        # update
        },
    ## ------------------------------------------------------------------------
    # Time series-related 
    ## ------------------------------------------------------------------------
    'time_series': {
        # asfreq
        # asof
        # shift
        # first_valid_index
        # last_valid_index
        # resample
        # tz_convert
        # tz_localize
        # at_time
        # between_time
        # tshift
        # slice_shift
        },
    ## ------------------------------------------------------------------------
    # Datetime properties 
    ## ------------------------------------------------------------------------
    'datetime_properties': {
        'dt.date': _.dt.date                         >> Elwise(postgresql = 'not_impl'), # TODO: all 3, not pandas objects
        'dt.time': _.dt.time                             >> Elwise(postgresql = 'not_impl'),
        'dt.timetz': _.dt.timetz                         >> Elwise(postgresql = 'not_impl'),
        'dt.year': _.dt.year                             >> Elwise(sql_type = 'float'),
        'dt.month': _.dt.month                           >> Elwise(sql_type = 'float'),
        'dt.day': _.dt.day                               >> Elwise(sql_type = 'float'),
        'dt.hour': _.dt.hour                             >> Elwise(sql_type = 'float'),
        'dt.minute': _.dt.minute                         >> Elwise(sql_type = 'float'),
        'dt.second': _.dt.second                         >> Elwise(sql_type = 'float'),
        'dt.microsecond': _.dt.microsecond               >> Elwise(sql_type = 'float', postgresql = 'xfail'),
        'dt.nanosecond': _.dt.nanosecond                 >> Elwise(postgresql = 'not_impl'),
        'dt.week': _.dt.week                             >> Elwise(sql_type = 'float'),
        'dt.weekofyear': _.dt.weekofyear                 >> Elwise(sql_type = 'float'),
        'dt.dayofweek': _.dt.dayofweek                   >> Elwise(sql_type = 'float'),
        'dt.weekday': _.dt.weekday                       >> Elwise(sql_type = 'float'),
        'dt.dayofyear': _.dt.dayofyear                   >> Elwise(sql_type = 'float'),
        'dt.quarter': _.dt.quarter                       >> Elwise(sql_type = 'float'),
        'dt.is_month_start': _.dt.is_month_start         >> Elwise(),
        'dt.is_month_end': _.dt.is_month_end             >> Elwise(),
        'dt.is_quarter_start': _.dt.is_quarter_start     >> Elwise(),
        'dt.is_quarter_end': _.dt.is_quarter_end         >> Elwise(postgresql = 'xfail'),
        'dt.is_year_start': _.dt.is_year_start           >> Elwise(),
        'dt.is_year_end': _.dt.is_year_end               >> Elwise(),
        'dt.is_leap_year': _.dt.is_leap_year             >> Elwise(postgresql = 'not_impl'),
        'dt.daysinmonth': _.dt.daysinmonth               >> Elwise(sql_type = 'float'),
        'dt.days_in_month': _.dt.days_in_month           >> Elwise(sql_type = 'float'),
        'dt.tz': _.dt.tz                                 >> Singleton(),
        'dt.freq': _.dt.freq                             >> Singleton(),
        },
    ## ------------------------------------------------------------------------
    # Datetime methods 
    ## ------------------------------------------------------------------------
    'datetime_methods': {
        'dt.to_period': _.dt.to_period('D')             >> Elwise(postgresql = 'xfail'),
        # dt.to_pydatetime                                              # TODO: datetime objects converted back to numpy?
        'dt.tz_localize': _.dt.tz_localize('UTC')       >> Elwise(postgresql = 'xfail'),
        # dt.tz_convert                                                 # TODO: need custom test
        'dt.normalize': _.dt.normalize()                >> Elwise(postgresql = 'xfail'),
        'dt.strftime': _.dt.strftime('%d')              >> Elwise(postgresql = 'xfail'),
        'dt.round': _.dt.round('D')                     >> Elwise(postgresql = 'xfail'),
        'dt.floor': _.dt.floor('D')                     >> Elwise(postgresql = 'xfail'),
        'dt.ceil': _.dt.ceil('D')                       >> Elwise(postgresql = 'xfail'),
        'dt.month_name': _.dt.month_name()              >> Elwise(postgresql = 'xfail'),
        'dt.day_name': _.dt.day_name()                  >> Elwise(postgresql = 'xfail'),
        },
    ## ------------------------------------------------------------------------
    # Period properties 
    ## ------------------------------------------------------------------------
    'period_properties': {
        # dt.qyear
        # dt.start_time
        # dt.end_time
        },
    ## ------------------------------------------------------------------------
    # Timedelta properties 
    ## ------------------------------------------------------------------------
    'timedelta_properties': {
        # dt.days
        # dt.seconds
        # dt.microseconds
        # dt.nanoseconds
        # dt.components
        },
    ## ------------------------------------------------------------------------
    # Timedelta methods 
    ## ------------------------------------------------------------------------
    'timedelta_methods': {
        # dt.to_pytimedelta
        # dt.total_seconds
        },
    ## ------------------------------------------------------------------------
    ## String methods
    ## ------------------------------------------------------------------------
    'string_methods': {
        'str.capitalize': _.str.capitalize()              >> Elwise(),
        #'str.casefold': _.str.casefold()                  >> Elwise(),   #TODO: introduced in v0.25.1
        # str.cat                                                         #TODO: can be Agg OR Elwise, others arg
        'str.center': _.str.center(3)                     >> Elwise(postgresql = 'not_impl'),
        'str.contains': _.str.contains('a')               >> Elwise(),
        'str.count': _.str.count('a')                     >> Elwise(postgresql = 'xfail'),
        # str.decode                                                      # TODO custom testing
        'str.encode': _.str.encode('utf-8')               >> Elwise(postgresql = 'xfail'),
        'str.endswith': _.str.endswith('a|b')             >> Elwise(postgresql = 'xfail'),
        #'str.extract': _.str.extract('(a)(b)')                           # TODO: returns DataFrame
        # str.extractall
        'str.find': _.str.find('a|c')                     >> Elwise(postgresql = 'xfail'),
        'str.findall': _.str.findall('a|c')               >> Elwise(postgresql = 'xfail'),
        # str.get                                                         # TODO: custom test
        # str.index                                                       # TODO: custom test
        # str.join                                                        # TODO: custom test
        'str.len': _.str.len()                            >> Elwise(),
        'str.ljust': _.str.ljust(5)                       >> Elwise(postgresql = 'xfail'), # pg formatstr function
        'str.lower': _.str.lower()                        >> Elwise(),
        'str.lstrip': _.str.lstrip()                      >> Elwise(),
        'str.match': _.str.match('a|c')                   >> Elwise(postgresql = 'xfail'),
        # str.normalize
        'str.pad': _.str.pad(5)                           >> Elwise(postgresql = 'xfail'),
        # str.partition
        # str.repeat
        'str.replace': _.str.replace('a|b', 'c')          >> Elwise(postgresql = 'xfail'),
        'str.rfind': _.str.rfind('a')                     >> Elwise(postgresql = 'xfail'),
        # str.rindex
        'str.rjust': _.str.rjust(5)                       >> Elwise(postgresql = 'xfail'),
        # str.rpartition
        'str.rstrip': _.str.rstrip()                      >> Elwise(),
        'str.slice': _.str.slice(step = 2)                >> Elwise(postgresql = 'xfail'),
        'str.slice_replace': _.str.slice_replace(2, repl = 'x')   >> Elwise(postgresql = 'xfail'),
        'str.split': _.str.split('a|b')                   >> Elwise(postgresql = 'xfail'),
        'str.rsplit': _.str.rsplit('a|b', n = 1)          >> Elwise(postgresql = 'xfail'),
        'str.startswith': _.str.startswith('a|b')         >> Elwise(),
        'str.strip': _.str.strip()                        >> Elwise(),
        'str.swapcase': _.str.swapcase()                  >> Elwise(postgresql = 'xfail'),
        'str.title': _.str.title()                        >> Elwise(),
        # str.translate
        'str.upper': _.str.upper()                        >> Elwise(),
        'str.wrap': _.str.wrap(2)                         >> Elwise(postgresql = 'xfail'),
        # str.zfill
        'str.isalnum': _.str.isalnum()                    >> Elwise(postgresql = 'xfail'),
        'str.isalpha': _.str.isalpha()                    >> Elwise(postgresql = 'xfail'),
        'str.isdigit': _.str.isdigit()                    >> Elwise(postgresql = 'xfail'),
        'str.isspace': _.str.isspace()                    >> Elwise(postgresql = 'xfail'),
        'str.islower': _.str.islower()                    >> Elwise(postgresql = 'xfail'),
        'str.isupper': _.str.isupper()                    >> Elwise(postgresql = 'xfail'),
        'str.istitle': _.str.istitle()                    >> Elwise(postgresql = 'xfail'),
        'str.isnumeric': _.str.isnumeric()                >> Elwise(postgresql = 'xfail'),
        'str.isdecimal': _.str.isdecimal()                >> Elwise(postgresql = 'xfail'),
        # str.get_dummies
        },
    'categories': {
        # cat.categories
        # cat.ordered
        # cat.codes
        # cat.rename_categories
        # cat.reorder_categories
        # cat.add_categories
        # cat.remove_categories
        # cat.remove_unused_categories
        # cat.set_categories
        # cat.as_ordered
        # cat.as_unordered
        },
    'sparse': {
        # sparse
        # sparse.npoints
        # sparse.density
        # sparse.fill_value
        # sparse.sp_values
        # sparse.from_coo
        # sparse.to_coo
        },
    ## ------------------------------------------------------------------------
    # Plotting 
    ## ------------------------------------------------------------------------
    # plot
    # plot.area
    # plot.bar
    # plot.barh
    # plot.box
    # plot.density
    # plot.hist
    # plot.kde
    # plot.line
    # plot.pie
    # hist
    ## ------------------------------------------------------------------------
    # Serialization / IO / conversion 
    ## ------------------------------------------------------------------------
    # to_pickle
    # to_csv
    # to_dict
    # to_excel
    # to_frame
    # to_xarray
    # to_hdf
    # to_sql
    # to_msgpack
    # to_json
    # to_dense
    # to_string
    # to_clipboard
    # to_latex
    }

from siuba.spec.utils import get_type_info
import itertools

funcs_stripped = { 
        section_name: { k: strip_symbolic(v) for k,v in section.items()}
          for section_name, section in funcs.items()
          } 


all_funcs = dict(itertools.chain(*[x.items() for x in funcs_stripped.values()]))


# Get spec =====
nested_spec = {}
for category, call_dict in funcs_stripped.items():
    nested_spec[category] = d = {}
    for name, call in call_dict.items():
        d[name] = get_type_info(call)
        d[name]['category'] = category


spec = dict(itertools.chain(*iter(d.items() for d in nested_spec.values())))

if __name__ == "__main__":
    from siuba.spec.utils import dump_spec
    import sys
    from pathlib import Path
    path_spec = Path(sys.argv[0]).parent / 'series.yml'
    
    with open(str(path_spec), 'w') as f:
        dump_spec(nested_spec, f)



from siuba import _, transmute, group_by

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql

DATA = data_frame(a = [1,2,3], b = [9,8,7])

@pytest.fixture(scope = "module")
def dfs(backend):
    return backend.load_df(DATA)


def test_transmute_basic(dfs):
    assert_equal_query(dfs, transmute(x = _.a + 1), data_frame(x = [2,3,4]))

def test_transmute_grouped(dfs):
    assert_equal_query(
            dfs,
            group_by(_.a) >> transmute(x = _.a + 1),
            data_frame(a = [1,2,3], x = [2,3,4])
            )

@pytest.mark.skip("TODO (#111)")
def test_transmute_grouped_no_mutate_grouping(dfs):
    assert_equal_query(
            dfs,
            group_by(_.a) >> transmute(a = _.a + 1),
            data_frame(a = [1,2,3], x = [2,3,4])
            )



from siuba.dply.verbs import (
        singledispatch2,
        show_query, collect,
        simple_varname,
        select, VarList, var_select,
        mutate,
        transmute,
        filter,
        arrange, _call_strip_ascending,
        summarize,
        count,
        group_by, ungroup,
        case_when,
        join, left_join, right_join, inner_join, semi_join, anti_join,
        head,
        rename,
        distinct,
        if_else
        )
from .translate import sa_get_over_clauses, CustomOverClause, SqlColumn, SqlColumnAgg
from .utils import get_dialect_funcs, get_sql_classes

from sqlalchemy import sql
import sqlalchemy
from siuba.siu import Call, CallTreeLocal, str_to_getitem_call, Lazy, FunctionLookupError
# TODO: currently needed for select, but can we remove pandas?
from pandas import Series
import pandas as pd

from sqlalchemy.sql import schema


# TODO:
#   - distinct
#   - annotate functions using sel.prefix_with("\n/*<Mutate Statement>*/\n") ?


# Helpers ---------------------------------------------------------------------

class SqlFunctionLookupError(FunctionLookupError): pass


class CallListener:
    """Generic listener. Each exit is called on a node's copy."""
    def enter(self, node):
        args, kwargs = node.map_subcalls(self.enter)

        return self.exit(node.__class__(node.func, *args, **kwargs))

    def exit(self, node):
        return node


class WindowReplacer(CallListener):
    """Call tree listener.

    Produces 2 important behaviors via the enter method:
      - returns evaluated sql call expression, with labels on all window expressions.
      - stores all labeled window expressions via the windows property.

    TODO: could replace with a sqlalchemy transformer
    """

    def __init__(self, columns, group_by, order_by, window_cte = None):
        self.columns = columns
        self.group_by = group_by
        self.order_by = order_by
        self.window_cte = window_cte
        self.windows = []

    def exit(self, node):
        col_expr = node(self.columns)

        if not isinstance(col_expr, sql.elements.ClauseElement):
            return col_expr

        over_clauses = [x for x in sa_get_over_clauses(col_expr) if isinstance(x, CustomOverClause)]

        # put groupings and orderings onto custom over clauses
        for over in over_clauses:
            # TODO: shouldn't mutate these over clauses
            group_by = sql.elements.ClauseList(
                    *[self.columns[name] for name in self.group_by]
                    )
            order_by = sql.elements.ClauseList(
                    *_create_order_by_clause(self.columns, *self.order_by)
                    )

            over.set_over(group_by, order_by)

        if len(over_clauses) and self.window_cte is not None:
            label = col_expr.label(None)
            self.windows.append(label)

            # optionally put into CTE, and return its resulting column
            self.window_cte.append_column(label)
            win_col = self.window_cte.c.values()[-1]
            return win_col
                
        return col_expr


def track_call_windows(call, columns, group_by, order_by, window_cte = None):
    listener = WindowReplacer(columns, group_by, order_by, window_cte)
    col = listener.enter(call)
    return col, listener.windows


def lift_inner_cols(tbl):
    cols = list(tbl.inner_columns)
    data = {col.key: col for col in cols}

    return sql.base.ImmutableColumnCollection(data, cols)

def col_expr_requires_cte(call, sel, is_mutate = False):
    """Return whether a variable assignment needs a CTE"""

    call_vars = set(call.op_vars(attr_calls = False))

    columns = lift_inner_cols(sel)
    sel_labs = set(k for k,v in columns.items() if isinstance(v, sql.elements.Label))

    # I use the acronym fwg sol (frog soul) to remember sql clause eval order
    # from, where, group by, select, order by, limit
    # group clause evaluated before select clause, so not issue for mutate
    group_needs_cte = not is_mutate and len(sel._group_by_clause)
    
    return (   group_needs_cte
            or len(sel._order_by_clause)
            or not sel_labs.isdisjoint(call_vars)
            )

def get_missing_columns(call, columns):
    missing_cols = set(call.op_vars(attr_calls = False)) - set(columns.keys())
    return missing_cols

def compile_el(tbl, el):
    compiled = el.compile(
         dialect = tbl.source.dialect,
         compile_kwargs = {"literal_binds": True}
    )
    return compiled

# Misc utilities --------------------------------------------------------------

def ordered_union(x, y):
    dx = {el: True for el in x}
    dy = {el: True for el in y}

    return tuple({**dx, **dy})




# Table -----------------------------------------------------------------------

class LazyTbl:
    def __init__(
            self, source, tbl, columns = None,
            ops = None, group_by = tuple(), order_by = tuple(), funcs = None,
            rm_attr = ('str', 'dt'), call_sub_attr = ('dt', 'str'),
            dispatch_cls = None,
            result_cls = sql.elements.ClauseElement
            ):
        """Create a representation of a SQL table.

        Args:
            source: a sqlalchemy.Engine or sqlalchemy.Connection instance.
            tbl: table of form 'schema_name.table_name', 'table_name', or sqlalchemy.Table.
            columns: if specified, a listlike of column names.

        Examples
        --------

        ::
            from sqlalchemy import create_engine
            from siuba.data import mtcars

            # create database and table
            engine = create_engine("sqlite:///:memory:")
            mtcars.to_sql('mtcars', engine)

            tbl_mtcars = LazyTbl(engine, 'mtcars')
            
        """
        
        # connection and dialect specific functions
        self.source = sqlalchemy.create_engine(source) if isinstance(source, str) else source

        dialect = self.source.dialect.name
        self.funcs = get_dialect_funcs(dialect) if funcs is None else funcs
        self.dispatch_cls = get_sql_classes(dialect) if dispatch_cls is None else dispatch_cls
        self.result_cls = result_cls

        self.tbl = self._create_table(tbl, columns, self.source)

        # important states the query can be in (e.g. grouped)
        self.ops = [sql.Select([self.tbl])] if ops is None else ops

        self.group_by = group_by
        self.order_by = order_by

        # customizations to allow interop with pandas (e.g. handle dt methods)
        self.rm_attr = rm_attr
        self.call_sub_attr = call_sub_attr

        self.result_cls = result_cls


    def append_op(self, op, **kwargs):
        cpy = self.copy(**kwargs)
        cpy.ops = cpy.ops + [op]
        return cpy

    def copy(self, **kwargs):
        return self.__class__(**{**self.__dict__, **kwargs})

    def shape_call(
            self,
            call, window = True, str_accessors = False,
            verb_name = None, arg_name = None,
            ):
        # TODO: error if mutate receives a literal value?
        # TODO: dispatch_cls currently unused
        if str_accessors and isinstance(call, str):
            # verbs that can use strings as accessors, like group_by, or
            # arrange, need to convert those strings into a getitem call
            return str_to_get_item_call(call)
        elif isinstance(call, sql.elements.ColumnClause):
            return Lazy(call)
        elif not isinstance(call, Call):
            # verbs that use literal strings, need to convert them to a call
            # that returns a sqlalchemy "literal" object
            return Lazy(sql.literal(call))

        # set up locals funcs dict
        f_dict1 = self.funcs['scalar']
        f_dict2 = self.funcs['window' if window else 'aggregate']

        funcs = {**f_dict1, **f_dict2}

        # determine dispatch class
        cls_name = 'window' if window else 'aggregate'
        dispatch_cls = self.dispatch_cls[cls_name]

        call_shaper = CallTreeLocal(
                funcs,
                call_sub_attr = self.call_sub_attr,
                dispatch_cls = dispatch_cls,
                result_cls = self.result_cls
                )

        # raise informative error message if missing translation
        try:
            return call_shaper.enter(call)
        except FunctionLookupError as err:
            raise SqlFunctionLookupError.from_verb(
                    verb_name or "Unknown",
                    arg_name or "Unknown",
                    err,
                    short = True
                    )


    def track_call_windows(self, call, columns = None, window_cte = None):
        """Returns tuple of (new column expression, list of window exprs)"""

        columns = self.last_op.columns if columns is None else columns
        return track_call_windows(call, columns, self.group_by, self.order_by, window_cte)

    def get_ordered_col_names(self):
        ungrouped = [k for k in self.last_op.columns.keys() if k not in self.group_by]
        return list(self.group_by) + ungrouped

    @property
    def last_op(self):
        return self.ops[-1] if len(self.ops) else None

    @staticmethod
    def _create_table(tbl, columns = None, source = None):
        """Return a sqlalchemy.Table, autoloading column info if needed. 

        Arguments:
            tbl: a sqlalchemy.Table or string of form 'table_name' or 'schema_name.table_name'.
            columns: a tuple of column names for the table. Overrides source argument.
            source: a sqlalchemy engine, used to autoload columns.

        """
        if isinstance(tbl, sqlalchemy.Table):
            return tbl
        
        if not isinstance(tbl, str):
            raise ValueError("tbl must be a sqlalchemy Table or string, but was %s" %type(tbl))

        if columns is None and source is None:
            raise ValueError("One of columns or source must be specified")

        schema, table_name = tbl.split('.') if '.' in tbl else [None, tbl]

        columns = map(sqlalchemy.Column, columns) if columns is not None else tuple()

        return sqlalchemy.Table(
                table_name,
                sqlalchemy.MetaData(bind = source),
                *columns,
                schema = schema,
                autoload_with = source if not columns else None
                )

    def _get_preview(self):
        # need to make prev op a cte, so we don't override any previous limit
        new_sel = sql.select([self.last_op.alias()]).limit(5)
        tbl_small = self.append_op(new_sel)
        return collect(tbl_small)

    def __repr__(self):
        template = (
                "# Source: lazy query\n"
                "# DB Conn: {}\n"
                "# Preview:\n{}\n"
                "# .. may have more rows"
                )

        return template.format(repr(self.source.engine), repr(self._get_preview()))

    def _repr_html_(self):
        template = (
                "<div>"
                "<pre>"
                "# Source: lazy query\n"
                "# DB Conn: {}\n"
                "# Preview:\n"
                "</pre>"
                "{}"
                "<p># .. may have more rows</p>"
                "</div>"
                )

        data = self._get_preview()
        html_data = getattr(data, '_repr_html_', lambda: repr(data))()
        return template.format(self.source.engine, html_data)


def _repr_grouped_df_html_(self):
    return "<div><p>(grouped data frame)</p>" + self._selected_obj._repr_html_() + "</div>"



# Main Funcs 
# =============================================================================

# sql raw --------------

sql_raw = sql.literal_column

# show query -----------

from sqlalchemy.ext.compiler import compiles, deregister
from contextlib import contextmanager

@contextmanager
def use_simple_names():
    get_col_name = lambda el, *args, **kwargs: str(el.element.name)
    try:
        yield compiles(sql.compiler._CompileLabel)(get_col_name)
    except:
        pass
    finally:
        deregister(sql.compiler._CompileLabel)

@show_query.register(LazyTbl)
def _show_query(tbl, simplify = False):
    query = tbl.last_op #if not simplify else 
    compile_query = lambda: query.compile(
                dialect = tbl.source.dialect,
                compile_kwargs = {"literal_binds": True}
            )


    if simplify:
        # try to strip table names and labels where uneccessary
        with use_simple_names():
            print(compile_query())
    else:
        # use a much more verbose query
        print(compile_query())

    return tbl

# collect ----------

@collect.register(LazyTbl)
def _collect(__data, as_df = True):
    # TODO: maybe remove as_df options, always return dataframe
    # normally can just pass the sql objects to execute, but for some reason
    # psycopg2 completes about incomplete template.
    # see https://stackoverflow.com/a/47193568/1144523
    query = __data.last_op
    compiled = query.compile(
        dialect = __data.source.dialect,
        compile_kwargs = {"literal_binds": True}
    )
    if as_df:
        return pd.read_sql(compiled, __data.source)

    return __data.source.execute(compiled).fetchall()


@select.register(LazyTbl)
def _select(__data, *args, **kwargs):
    # see https://stackoverflow.com/questions/25914329/rearrange-columns-in-sqlalchemy-select-object
    if kwargs:
        raise NotImplementedError(
                "Using kwargs in select not currently supported. "
                "Use _.newname == _.oldname instead"
                )
    last_op = __data.last_op
    columns = {c.key: c for c in last_op.inner_columns}

    # same as for DataFrame
    colnames = Series(list(columns))
    vl = VarList()
    evaluated = (arg(vl) if callable(arg) else arg for arg in args)
    od = var_select(colnames, *evaluated)

    col_list = []
    for k,v in od.items():
        col = columns[k]
        col_list.append(col if v is None else col.label(v))

    return __data.append_op(last_op.with_only_columns(col_list))



@filter.register(LazyTbl)
def _filter(__data, *args, **kwargs):
    # TODO: aggregate funcs
    # Note: currently always produces 2 additional select statements,
    #       1 for window/aggs, and 1 for the where clause
    sel = __data.last_op.alias()
    win_sel = sql.select([sel], from_obj = sel)

    conds = []
    windows = []
    for ii, arg in enumerate(args):
        if isinstance(arg, Call):
            new_call = __data.shape_call(arg, verb_name = "Filter", arg_name = ii)
            #var_cols = new_call.op_vars(attr_calls = False)

            col_expr, win_cols = __data.track_call_windows(
                    new_call,
                    sel.columns,
                    window_cte = win_sel
                    )

            conds.append(col_expr)
        else:
            conds.append(arg)

    bool_clause = sql.and_(*conds)

    # move non-window functions to refer to win_sel clause (not the innermost)
    win_alias = win_sel.alias()
    bool_clause = sql.util.ClauseAdapter(win_alias).traverse(bool_clause)

    
    orig_cols = [win_alias.columns[k] for k in __data.get_ordered_col_names()]
    filt_sel = sql.select(orig_cols, from_obj = win_alias, whereclause = bool_clause)
    return __data.append_op(filt_sel)


@mutate.register(LazyTbl)
def _mutate(__data, **kwargs):
    # Cases
    #  - work with group by
    #  - window functions
    # TODO: verify it can follow a renaming select

    # track labeled columns in set
    sel = __data.last_op

    # evaluate each call
    for colname, func in kwargs.items():
        # keep set of columns labeled (aliased) in this select statement
        # need to use inner cols, since sel.columns uses ColumnClause, not Label
        labs = set(k for k,v in lift_inner_cols(sel).items() if isinstance(v, sql.elements.Label))
        new_call = __data.shape_call(func, verb_name = "Mutate", arg_name = colname)

        sel = _mutate_select(sel, colname, new_call, labs, __data)

    return __data.append_op(sel)


def _mutate_select(sel, colname, func, labs, __data):
    """Return select statement containing a new column, func expr as colname.
    
    Note: since a func can refer to previous columns generated by mutate, this
    function handles whether to add a column to the existing select statement,
    or to use it as a subquery.
    """
    replace_col = colname in sel.columns
    # Call objects let us check whether column expr used a derived column
    # e.g. SELECT a as b, b + 1 as c raises an error in SQL, so need subquery
    if not col_expr_requires_cte(func, sel, is_mutate = True):
        # New column may be able to modify existing select
        columns = lift_inner_cols(sel)

    else:
        # anything else requires a subquery
        cte = sel.alias(None)
        columns = cte.columns
        sel = sql.select([cte], from_obj = cte)

    # evaluate call expr on columns, making sure to use group vars
    new_col, windows = __data.track_call_windows(func, columns)

    # replacing an existing column, so strip it from select statement
    if replace_col:
        replaced = {**columns}
        replaced[colname] = new_col.label(colname)
        return sel.with_only_columns(list(replaced.values()))

    return sel.column(new_col.label(colname))


@transmute.register(LazyTbl)
def _transmute(__data, **kwargs):
    # will use mutate, then select some cols
    f_mutate = mutate.registry[type(__data)]

    # transmute keeps grouping cols, and any defined in kwargs
    cols_to_keep = ordered_union(__data.group_by, kwargs)

    sel = f_mutate(__data, **kwargs).last_op

    columns = lift_inner_cols(sel)
    sel_stripped = sel.with_only_columns([columns[k] for k in cols_to_keep])

    return __data.append_op(sel_stripped)


@arrange.register(LazyTbl)
def _arrange(__data, *args):
    last_op = __data.last_op
    cols = lift_inner_cols(last_op)

    
    new_calls = []
    for ii, expr in enumerate(args):
        if callable(expr):
            res = __data.shape_call(
                    expr, window = False,
                    verb_name = "Arrange", arg_name = ii
                    )

        else:
            res = expr

        new_calls.append(res)

    sort_cols = _create_order_by_clause(cols, *new_calls)

    order_by = __data.order_by + tuple(new_calls)
    return __data.append_op(last_op.order_by(*sort_cols), order_by = order_by)


# TODO: consolidate / pull expr handling funcs into own file?
def _create_order_by_clause(columns, *args):
    sort_cols = []
    for arg in args:
        # simple named column
        if isinstance(arg, str):
            sort_cols.append(columns[arg])
        # an expression
        elif callable(arg):
            #f, asc = _call_strip_ascending(arg)
            #col_op = f(cols) if asc else f(cols).desc()
            col_op = arg(columns)
            sort_cols.append(col_op)
        else:
            raise NotImplementedError("Must be string or callable")

    return sort_cols



@count.register(LazyTbl)
def _count(__data, *args, sort = False, wt = None, **kwargs):
    # TODO: if already col named n, use name nn, etc.. get logic from tidy.py
    if kwargs:
        raise NotImplementedError("TODO")

    if wt is not None:
        raise NotImplementedError("TODO")

    # similar to filter verb, we need two select statements,
    # an inner one for derived cols, and outer to group by them
    sel = __data.last_op.alias()
    sel_inner = sql.select([sel], from_obj = sel)

    # inner select ----
    # holds any mutation style columns
    group_cols = []
    for arg in args:
        col_name = simple_varname(arg)
        if col_name is None:
            # evaluate call
            col_expr = arg(sel.columns) if callable(arg) else arg

            # compile, so we can use the expr as its name (e.g. "id + 1")
            col_name = str(compile_el(__data, col_expr))
            label = col_expr.label(col_name)
            sel_inner.append_column(label)

        group_cols.append(col_name)

    # outer select ----
    # holds selected columns and tally (n)
    sel_inner_cte = sel_inner.alias()
    inner_cols = sel_inner_cte.columns
    sel_outer = sql.select(from_obj = sel_inner_cte)

    # apply any group vars from a group_by verb call first
    prev_group_cols = [inner_cols[k] for k in __data.group_by]
    if prev_group_cols:
        sel_outer.append_group_by(*prev_group_cols)
        sel_outer.append_column(*prev_group_cols)

    # now any defined in the count verb call
    for k in group_cols:
        sel_outer.append_group_by(inner_cols[k])
        sel_outer.append_column(inner_cols[k])

    sel_outer.append_column(sql.functions.count().label("n"))

    return __data.append_op(sel_outer)
    


@summarize.register(LazyTbl)
def _summarize(__data, **kwargs):
    # https://stackoverflow.com/questions/14754994/why-is-sqlalchemy-count-much-slower-than-the-raw-query
    # what if windowed mutate or filter has been done? 
    #   - filter is fine, since it uses a CTE
    #   - need to detect any window functions...
    sel = __data.last_op._clone()

    new_calls = {}
    for k, expr in kwargs.items():
        new_calls[k] = __data.shape_call(
                expr, window = False,
                verb_name = "Summarize", arg_name = k
                )

    needs_cte = [col_expr_requires_cte(call, sel) for call in new_calls.values()]

    # create select statement ----
    if any(needs_cte):
        # need a cte, due to alias cols or existing group by
        # current select stmt has group by clause, so need to make it subquery
        cte = sel.alias()
        columns = cte.columns
        sel = sql.select(from_obj = cte)
    else:
        # otherwise, can alter the existing select statement
        columns = lift_inner_cols(sel)
        old_froms = sel.froms

        sel = sel.with_only_columns([])
        sel.append_from(*old_froms)

    # add group by columns ----
    group_cols = [columns[k] for k in __data.group_by]
    sel.append_group_by(*group_cols)
    for col in group_cols:
        sel.append_column(col)

    # add each aggregate column ----
    # TODO: can't do summarize(b = mean(a), c = b + mean(a))
    #       since difficult for c to refer to agg and unagg cols in SQL
    for k, expr in new_calls.items():
        missing_cols = get_missing_columns(expr, columns)
        if missing_cols:
            raise NotImplementedError(
                    "Summarize cannot find the following columns: %s. "
                    "Note that it cannot refer to variables defined earlier in the "
                    "same summarize call." % missing_cols
                    )

        col = expr(columns).label(k)

        sel.append_column(col)

    new_data = __data.append_op(sel, group_by = tuple(), order_by = tuple())
    return new_data


@group_by.register(LazyTbl)
def _group_by(__data, *args, add = False, **kwargs):
    if kwargs:
        data = mutate(__data, **kwargs)
    else:
        data = __data

    cols = data.last_op.columns

    # put kwarg grouping vars last, so similar order to function call
    groups =  tuple(simple_varname(arg) for arg in args) + tuple(kwargs)
    if None in groups:
        raise NotImplementedError("Complex expressions not supported in sql group_by")

    unmatched = set(groups) - set(cols.keys())
    if unmatched:
        raise KeyError("group_by specifies columns missing from table: %s" %unmatched)

    if add:
        groups = ordered_union(data.group_by, groups)

    return data.copy(group_by = groups)


@ungroup.register(LazyTbl)
def _ungroup(__data):
    return __data.copy(group_by = tuple())


@case_when.register(sql.base.ImmutableColumnCollection)
def _case_when(__data, cases):
    # TODO: will need listener to enter case statements, to handle when they use windows
    if isinstance(cases, Call):
        cases = cases(__data)

    whens = []
    case_items = list(cases.items())
    n_items = len(case_items)

    else_val = None
    for ii, (expr, val) in enumerate(case_items):
        # handle where val is a column expr
        if callable(val):
            val = val(__data)

        # handle when expressions
        if ii+1 == n_items and expr is True:
            else_val = val
        elif callable(expr):
            whens.append((expr(__data), val))
        else:
            whens.append((expr, val))

    return sql.case(whens, else_ = else_val)
        

# Join ------------------------------------------------------------------------

from collections.abc import Mapping

def _joined_cols(left_cols, right_cols, on_keys, full = False):
    """Return labeled columns, according to selection rules for joins.

    Rules:
        1. For join keys, keep left table's column
        2. When keys have the same labels, add suffix
    """
    # TODO: remove sets, so uses stable ordering
    # when left and right cols have same name, suffix with _x / _y
    keep_right = set(right_cols.keys()) - set(on_keys.values())
    shared_labs = set(left_cols.keys()).intersection(keep_right)

    right_cols_no_keys = {k: right_cols[k] for k in keep_right}

    # for an outer join, have key columns coalesce values
    if full:
        left_cols = {**left_cols}
        for lk, rk in on_keys.items():
            col = sql.functions.coalesce(left_cols[lk], right_cols[rk])
            left_cols[lk] = col.label(lk)

    # create labels ----
    l_labs = _relabeled_cols(left_cols, shared_labs, "_x")
    r_labs = _relabeled_cols(right_cols_no_keys, shared_labs, "_y")

    return l_labs + r_labs
    


def _relabeled_cols(columns, keys, suffix):
    # add a suffix to all columns with names in keys
    cols = []
    for k, v in columns.items():
        new_col = v.label(k + str(suffix)) if k in keys else v
        cols.append(new_col)
    return cols


@join.register(LazyTbl)
def _join(left, right, on = None, how = "inner"):
    # Needs to be on the table, not the select
    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on)
    how = _validate_join_arg_how(how)
    
    if how == "right":
        # switch joins, since sqlalchemy doesn't have right join arg
        # see https://stackoverflow.com/q/11400307/1144523
        left_sel, right_sel = right_sel, left_sel

    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create join ----
    join = left_sel.join(
            right_sel,
            onclause = bool_clause,
            isouter = how != "inner",
            full = how == "full"
            )
    
    # note, shared_keys assumes on is a mapping...
    shared_keys = [k for k,v in on.items() if k == v]
    labeled_cols = _joined_cols(
            left_sel.columns,
            right_sel.columns,
            on_keys = on,
            full = how == "full"
            )

    sel = sql.select(labeled_cols, from_obj = join)
    return left.append_op(sel)


@semi_join.register(LazyTbl)
def _semi_join(left, right = None, on = None):

    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on)
    
    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create inner join ----
    exists_clause = sql.select(
            [sql.literal(1)],
            from_obj = right_sel,
            whereclause = bool_clause
            )

    # only keep left hand select's columns ----
    sel = sql.select(
            left_sel.columns,
            from_obj = left_sel,
            whereclause = sql.exists(exists_clause)
            )

    return left.append_op(sel)


@anti_join.register(LazyTbl)
def _anti_join(left, right = None, on = None):
    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on)
    
    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create inner join ----
    not_exists = ~sql.exists([1], from_obj = right_sel).where(bool_clause)
    sel = sql.select(left_sel.columns, from_obj = left_sel).where(not_exists)
    return left.append_op(sel)
       

def _validate_join_arg_on(on):
    if on is None:
        raise NotImplementedError("on arg must currently be dict")
    elif isinstance(on, str):
        on = {on: on}
    elif isinstance(on, (list, tuple)):
        on = dict(zip(on, on))

    if not isinstance(on, Mapping):
        raise TypeError("on must be a Mapping (e.g. dict)")

    return on

def _validate_join_arg_how(how):
    how_options = ("inner", "left", "right", "full")
    if how not in how_options:
        raise ValueError("how argument needs to be one of %s" %how_options)
    
    return how

def _create_join_conds(left_sel, right_sel, on):
    left_cols  = left_sel.columns  #lift_inner_cols(left_sel)
    right_cols = right_sel.columns #lift_inner_cols(right_sel)

    conds = []
    for l, r in on.items():
        col_expr = left_cols[l] == right_cols[r]
        conds.append(col_expr)
        
    return sql.and_(*conds)
    

# Head ------------------------------------------------------------------------

@head.register(LazyTbl)
def _head(__data, n = 5):
    sel = __data.last_op
    
    return __data.append_op(sel.limit(n))


# Rename ----------------------------------------------------------------------

@rename.register(LazyTbl)
def _rename(__data, **kwargs):
    sel = __data.last_op
    columns = lift_inner_cols(sel)

    # old_keys uses dict as ordered set
    old_to_new = {simple_varname(v):k for k,v in kwargs.items()}
    
    if None in old_to_new:
        raise KeyError("positional arguments must be simple column, "
                        "e.g. _.colname or _['colname']"
                        )

    labs = [c.label(old_to_new[k]) if k in old_to_new else c for k,c in columns.items()]

    new_sel = sel.with_only_columns(labs)

    return __data.append_op(new_sel)


# Distinct --------------------------------------------------------------------

@distinct.register(LazyTbl)
def _distinct(__data, *args, _keep_all = False, **kwargs):
    if (args or kwargs) and _keep_all:
        raise NotImplementedError("Distinct with variables specified in sql requires _keep_all = False")
    
    inner_sel = mutate(__data, **kwargs).last_op if kwargs else __data.last_op

    # TODO: this is copied from the df distinct version
    # cols dict below is used as ordered set
    cols = {simple_varname(x): True for x in args}
    cols.update(kwargs)

    if None in cols:
        raise KeyError("positional arguments must be simple column, "
                        "e.g. _.colname or _['colname']"
                        )

    # use all columns by default
    if not cols:
        cols = list(inner_sel.columns.keys())

    if not len(inner_sel._order_by_clause):
        # select distinct has to include any columns in the order by clause,
        # so can only safely modify existing statement when there's no order by
        sel_cols = lift_inner_cols(inner_sel)
        distinct_cols = [sel_cols[k] for k in cols]
        sel = inner_sel.with_only_columns(distinct_cols).distinct()
    else:
        # fallback to cte
        cte = inner_sel.alias()
        distinct_cols = [cte.columns[k] for k in cols]
        sel = sql.select(distinct_cols, from_obj = cte).distinct()

    return __data.append_op(sel)

    
# if_else ---------------------------------------------------------------------

@if_else.register(sql.elements.ColumnElement)
def _if_else(cond, true_vals, false_vals):
    whens = [(cond, true_vals)]
    return sql.case(whens, else_ = false_vals)





"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-group_by.R
"""
    
from siuba import _, group_by, ungroup, summarize
from siuba.dply.vector import row_number, n

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, SqlBackend
from string import ascii_lowercase 

DATA = data_frame(x = [1,2,3], y = [9,8,7], g = ['a', 'a', 'b'])

@pytest.fixture(scope = "module")
def df(backend):
    if not isinstance(backend, SqlBackend):
        pytest.skip("TODO: generalize tests to pandas")
    return backend.load_df(DATA)


def test_group_by_no_add(df):
    gdf = group_by(df, _.x, _.y)
    assert gdf.group_by == ("x", "y")

def test_group_by_override(df):
    gdf = df >> group_by(_.x, _.y) >> group_by(_.g)
    assert gdf.group_by == ("g",)

def test_group_by_add(df):
    gdf = group_by(df, _.x) >> group_by(_.y, add = True)

    assert gdf.group_by == ("x", "y")

def test_group_by_ungroup(df):
    q1 = df >> group_by(_.g)
    assert q1.group_by == ("g",)

    q2 = q1 >> ungroup()
    assert q2.group_by == tuple()


@pytest.mark.skip("TODO: need to test / validate joins first")
def test_group_by_before_joins(df):
    assert False

def test_group_by_performs_mutate(df):
    assert_equal_query(
            df,
            group_by(z = _.x + _.y) >> summarize(n = n(_)),
            data_frame(z = 10, n = 3)
            )




from siuba.dply.string import str_c
from sqlalchemy.sql.elements import ClauseElement
from sqlalchemy import sql

@str_c.register(ClauseElement)
def _str_c_sql(x, *args, sep = "", collapse = None) -> ClauseElement:
    """
    Example:
    """

    if collapse is not None:
        raise NotImplementedError("For SQL, collapse argument of str_c not supported")

    if sep != "":
        raise NotImplementedError('For SQL, sep argument of str_c must be ""')

    return sql.func.concat(x, *args)



"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-distinct.R
"""
    
from siuba.sql import LazyTbl, collect
from siuba import _, distinct, group_by, summarize, arrange, mutate
from .helpers import assert_equal_query, backend_sql
import pandas as pd
import os

import pytest
from sqlalchemy import create_engine

DATA = pd.DataFrame({
    "x": [1,2,3,4,5],
    "y": [5,4,3,2,1]
    })

@pytest.fixture(scope = "module")
def df(backend):
    yield backend.load_df(DATA)

def test_distinct_no_args(df):
    assert_equal_query(df, distinct(), DATA.drop_duplicates())
    assert_equal_query(df, distinct(), distinct(DATA))

def test_distinct_one_arg(df):
    assert_equal_query(
            df,
            distinct(_.y),
            DATA.drop_duplicates(['y'])[['y']].reset_index(drop = True)
            )

    assert_equal_query(df, distinct(_.y), distinct(DATA, _.y))

@backend_sql
def test_distinct_keep_all_not_impl(backend, df):
    # TODO: should just mock LazyTbl
    with pytest.raises(NotImplementedError):
        distinct(df, _.y, _keep_all = True) >> collect()
    

@pytest.mark.xfail
def test_distinct_via_group_by(df):
    # NotImplemented
    assert False


def test_distinct_after_summarize(df):
    query = group_by(g = _.x) >> summarize(z = (_.y - _.y).min()) >> distinct(_.z)
    
    assert_equal_query(df, query, pd.DataFrame({'z': [0]}))

def test_distinct_after_arrange(df):
    query = arrange(_.x) >> distinct(_.y)

    assert_equal_query(df, query, pd.DataFrame({'y': [5,4,3,2,1]}))


def test_distinct_of_mutate_col(df):
    query = mutate(z = _.x + 1) >>  distinct(_.z)

    assert_equal_query(df, query, pd.DataFrame({'z': [2,3,4,5,6]}))


def test_distinct_kwargs(df):
    dst = DATA.drop_duplicates(['y', 'x']) \
              .rename(columns = {'x': 'a'}) \
              .reset_index(drop = True)[['y', 'a']]

    assert_equal_query(df, distinct(_.y, a = _.x), dst)


    




from siuba.siu import _, MetaArg, strip_symbolic
import itertools

def get_type_info(call):
    if call.func != "__rshift__":
        raise ValueError("Expected first expressions was >>")
        
    out = {}
    expr, result = call.args
    
    accessors = ['str', 'dt', 'cat']
    
    accessor = [ameth for ameth in accessors if ameth in expr.op_vars()] + [None]
            
    return dict(
        expr_series = expr,
        expr_frame = replace_meta_args(expr, _.x, _.y, _.z),
        accessor = accessor[0],
        result = result.to_dict(),
        is_property = expr.func == "__getattr__",
        data_arity = count_call_type(expr, MetaArg)
    )

def count_call_type(call, cls):
    """Counts the number of MetaArgs"""
    meta_count = 0
    def incr_meta_count(node):
        nonlocal meta_count
        if isinstance(node, cls): 
            meta_count += 1
        node.map_subcalls(incr_meta_count)
    
    incr_meta_count(call)
    
    return meta_count

def replace_meta_args(call, *args):
    replacements = map(strip_symbolic, args)
    
    def replace_next(node):
        # swap in first replacement for meta arg (when possible)
        if isinstance(node, MetaArg):
            return next(replacements)
        
        # otherwise, recurse into node
        args, kwargs = node.map_subcalls(replace_next)

        # return a copy of node
        return node.__class__(node.func, *args, **kwargs)

    return replace_next(call)

def dump_spec(spec, stream = None):
    import yaml
    # TODO: do this without mutating SafeDumper...
    yaml.SafeDumper.yaml_representers[None] = lambda self, data: \
        yaml.representer.SafeRepresenter.represent_str(
            self,
            str(data),
        )

    return yaml.safe_dump(spec, stream)



from siuba.siu import CallListener, strip_symbolic
from siuba.dply.verbs import singledispatch2, mutate
from pandas.core.groupby import DataFrameGroupBy
import pandas as pd
    

class AttrStripper(CallListener):
    """
    Returns a Call with attribute access for specified attribute names removed.

    Example:
        from siuba import _, strip_symbolic
        attr_strip = AttrStripper({'hp'})

        # returns Call: _ + _.rank()
        attr_strip.enter(strip_symbolic(_.hp + _.hp.rank()))
    """
            
    def __init__(self, rm_attr):
        self.rm_attr = rm_attr
        
    def exit___getattr__(self, node):
        obj, attr_name = node.args
        if attr_name in self.rm_attr:
            return obj
        
        return node

    @classmethod
    def run_on_node(cls, node, *args, **kwargs):
        inst = cls(*args, **kwargs)
        return inst.enter(node)


def _grouped_transform_or_apply(data, call):
    # get variable names used, ignoring called methods
    var_names = call.op_vars(attr_calls = False)
    if len(var_names) == 1:
        # when only a single column in expression, use more efficient transform
        col_name = list(var_names)[0]
        new_call = AttrStripper.run_on_node(call, var_names)

        return data[col_name].transform(new_call)
    else:
        # TODO: analyze n groups: if 1 value per group, is agg, if has all needed rows is mutate
        # otherwise, use generic apply, and handle two issues
        # issue 1: it can return a DataFrame
        res = data.apply(call)
        if isinstance(res, pd.DataFrame):
            raise ValueError()
        
        # issue 2: its index is group info + orig index, 
        group_names = [x.name for x in data.grouper.groupings]
        #group_names = res.index.names[:-1]
        gdf_index = data.obj.set_index(group_names)
        # put rows in original order
        return res[gdf_index.index].values
    

@singledispatch2(object)
def mutate2(__data, **kwargs):
    """Warning: this function is experimental"""
    # by default dispatch to regular mutate
    f_mutate = mutate.registry[type(__data)]
    return f_mutate(__data, **kwargs)


@mutate2.register(DataFrameGroupBy)
def _mutate2_group_by(__data, **kwargs):
    new_df = __data.obj.copy()
    crnt_df = __data
    for k, call in kwargs.items():
        new_df[k] = _grouped_transform_or_apply(crnt_df, call)
        crnt_df = new_df.groupby(__data.grouper.groupings)
    return crnt_df


# Tests =======================================================================

# op returns len of series
# op returns series with single value
# op returns scalar




# TODO: move siu.py into its own folder, add this to it (w/ Call Trees)
from typing import Any, Union, TypeVar
import inspect

def is_union(x):
    return getattr(x, '__origin__', None) is Union

def get_union_args(x):
    return getattr(x, '__args__', getattr(x, '__union_args__', None))

def is_flex_subclass(x, cls):
    if x is Any:
        return True
    
    return issubclass(x, cls)

def is_dispatch_func_subtype(f, input_cls, output_cls):
    """Returns whether a singledispatch function is subtype of some input and result class.
    
    A function is a subtype if it is input contravariant, and result covariant.
    
    Rules for evaluating return types <= output_cls:
        * Any always returns True
        * Union[A, B] returns true if either A or B are covariant
        * f(arg_name:TypeVar) -> TypeVar compares input_cls and output_cls
        * Simple return types checked via issubclass
        
    Args:
        input_cls - input class for first argument to function
        output_cls - output class for function result
    
    """
    sig = inspect.signature(f)
    # result annotation
    res_type = sig.return_annotation
    
    # first parameter annotation
    par0 = next(iter(sig.parameters.values()))
    par_type0 = par0.annotation
    
    # Case 1: no annotation
    if res_type is None:
        return False
    
    # Case 2: fancy annotations: Union, generic TypeVar
    if is_union(res_type) and get_union_args(res_type):
        sub_types = get_union_args(res_type)
        # passes if any unioned types are subclasses
        return any(map(lambda x: is_flex_subclass(x, output_cls), sub_types))
    elif isinstance(res_type, TypeVar):
        if res_type == par_type0:
            # using a generic type variable as first arg and result
            # return type must be covariant on input_cls
            return issubclass(input_cls, output_cls) and res_type.__covariant__
        else:
            raise TypeError("Generic type used as result, but not as first parameter")

    return is_flex_subclass(res_type, output_cls)
